Set Implicit Arguments.
Require Import Setoid.
Require Import BinPos.
Require Import BinInt.
Require Export BinList.
Require Export Ring_theory.

(* Some Tactic *)

(* to avoid conflicts with Coq booleans*)
Inductive RingBool : Set :=
 | RBtrue : RingBool
 | RBfalse : RingBool.
 		       
Ltac IN a l :=
 match l with
 | (cons a ?l) => constr:RBtrue
 | (cons _ ?l) => IN a l
 | (nil _) => constr:RBfalse
 end.

Ltac AddFv a l :=
 match (IN a l) with
 | RBtrue => constr:l
 | _ => constr:(cons a l)
 end.

Ltac Find_at a l :=
 match l with
 | (nil _) => constr:xH
 | (cons a _) => constr:xH
 | (cons _ ?l) => let p := Find_at a l in constr:(Psucc p)
 end.
		      
Ltac FV Cst add mul sub opp t fv :=
 let rec TFV t fv :=
  match t with
  | (add ?t1 ?t2) => 
    let fv1 := TFV t1 fv in let fv2 := TFV t2 fv1 in constr:fv2
  | (mul ?t1 ?t2) => 
    let fv1 := TFV t1 fv in let fv2 := TFV t2 fv1 in constr:fv2
  | (sub ?t1 ?t2) => 
    let fv1 := TFV t1 fv in let fv2 := TFV t2 fv1 in constr:fv2
  | (opp ?t1) => 
    let fv1 := TFV t1 fv in constr:fv1
  | _ =>  
    match Cst t with
    | RBfalse => let fv1 := AddFv t fv in constr:fv1
    | _ => constr:fv
    end
  end 
 in TFV t fv.

Module Type RING_POL_ENTRIES.
 
 (* Ring elements *)     
 Parameter R:Type.
 Parameter (rO rI : R) (radd rmul rsub: R->R->R) (ropp : R->R).
 Parameter req : R -> R -> Prop.
 
 (* Ring properties *)
 Parameter Rsth : Setoid_Theory R req.
 Parameter Reqe : ring_eq_ext radd rmul ropp req.
 Parameter ARth : almost_ring_theory rO rI radd rmul rsub ropp req.

 (* Coefficients *) 
 Parameter C: Type.
 Parameter (cO cI: C) (cadd cmul csub : C->C->C) (copp : C->C).
 Parameter ceqb : C->C->bool. 
 Parameter phi : C -> R.
 Parameter CRmorph : ring_morph rO rI radd rmul rsub ropp req
                                cO cI cadd cmul csub copp ceqb phi.

End RING_POL_ENTRIES.	   
	   
Module MakeRingPol (RE:RING_POL_ENTRIES).
 Import RE.

 (* R notations *)
 Notation "0" := rO. Notation "1" := rI.
 Notation "x + y" := (radd x y).  Notation "x * y " := (rmul x y).
 Notation "x - y " := (rsub x y). Notation "- x" := (ropp x).
 Notation "x == y" := (req x y).

 (* C notations *)		 
 Notation "x +! y" := (cadd x y). Notation "x *! y " := (cmul x y).
 Notation "x -! y " := (csub x y). Notation "-! x" := (copp x).
 Notation " x ?=! y" := (ceqb x y). Notation "[ x ]" := (phi x).

 (* Usefull tactics *)		  
  Add Setoid R req Rsth.
 Ltac rrefl := gen_reflexivity Rsth.
  Add Morphism radd : radd_ext.  exact Reqe.(Radd_ext). Qed.
  Add Morphism rmul : rmul_ext.  exact Reqe.(Rmul_ext). Qed.
  Add Morphism ropp : ropp_ext.  exact Reqe.(Ropp_ext). Qed.
  Add Morphism rsub : rsub_ext. exact (ARsub_ext Rsth Reqe ARth). Qed.
 Ltac rsimpl := gen_srewrite 0 1 radd rmul rsub ropp req Rsth Reqe ARth.
 Ltac add_push := gen_add_push radd Rsth Reqe ARth.
 Ltac mul_push := gen_mul_push rmul Rsth Reqe ARth.

 (* Definition des polynomes multivariables a coefficiant dans C :
    Le type [Pol] represente [X1 ... Xn].
    La representation utiliser est celle de Horner ou un
    polynome a [n] variable (C[X1 ... Xn]) est vu comme un polynome
    une variable (C[X1 ... Xn-1][Xn]). Avec des optimisations 
    pour rendre cela efficace et compact :

    - [Pc c] polynome constant : c*X1^0*...*Xn^0.

    - [Pinj j Q] est le polynome constant en Xn-j+1 ... Xn : 
      Q*Xn-j+1^0*...*Xn^0 avec Q \in C[X1 ... Xn-j].                 
                       
    - [PX P i Q] une forme de Horner optimis'e : P*Xn^i + Q
        avec P \in C[X1 ... Xn] et Q \in C[X1 ... X(n-1)]               *)

 Inductive Pol : Type :=
  | Pc : C -> Pol 
  | Pinj : positive -> Pol -> Pol                   
  | PX : Pol -> positive -> Pol -> Pol.

 Definition P0 := Pc cO.
 Definition P1 := Pc cI.
 
 Fixpoint Peq (P P' : Pol) {struct P'} : bool := 
  match P, P' with
  | Pc c, Pc c' => c ?=! c'
  | Pinj j Q, Pinj j' Q' => 
    match Pcompare j j' Eq with
    | Eq => Peq Q Q' 
    | _ => false 
    end
  | PX P i Q, PX P' i' Q' =>
    match Pcompare i i' Eq with
    | Eq => if Peq P P' then Peq Q Q' else false
    | _ => false
    end
  | _, _ => false
  end.

 Notation " P ?== P' " := (Peq P P').

 Definition mkPinj j P :=
  match P with 
  | Pc _ => P
  | Pinj j' Q => Pinj ((j + j'):positive) Q
  | _ => Pinj j P
  end.

 Definition mkPX P i Q := 
  match P with
  | Pc c => if c ?=! cO then mkPinj xH Q else PX P i Q
  | Pinj _ _ => PX P i Q
  | PX P' i' Q' => if Q' ?== P0 then PX P' (i' + i) Q else PX P i Q
  end.
 
 (** Inverse de l'addition *)
 
 Fixpoint Popp (P:Pol) : Pol := 
  match P with
  | Pc c => Pc (-! c)
  | Pinj j Q => Pinj j (Popp Q)
  | PX P i Q => PX (Popp P) i (Popp Q)
  end.
 
 Notation "-- P" := (Popp P).

 (** Addition et soustraction **)
 
 Fixpoint PaddC (P:Pol) (c:C) {struct P} : Pol :=
  match P with
  | Pc c1 => Pc (c1 +! c)
  | Pinj j Q => Pinj j (PaddC Q c)
  | PX P i Q => PX P i (PaddC Q c)
  end.

 Fixpoint PsubC (P:Pol) (c:C) {struct P} : Pol :=
  match P with
  | Pc c1 => Pc (c1 -! c)
  | Pinj j Q => Pinj j (PsubC Q c)
  | PX P i Q => PX P i (PsubC Q c)
  end.

 Section PopI.

  Variable Pop : Pol -> Pol -> Pol.
  Variable Q : Pol.

  Fixpoint PaddI (j:positive) (P:Pol){struct P} : Pol :=
   match P with
   | Pc c => mkPinj j (PaddC Q c)
   | Pinj j' Q' => 
     match ZPminus j' j with
     | Zpos k =>  mkPinj j (Pop (Pinj k Q') Q)
     | Z0 => mkPinj j (Pop Q' Q)
     | Zneg k => mkPinj j' (PaddI k Q')
     end
   | PX P i Q' => 
     match j with
     | xH => PX P i (Pop Q' Q)
     | xO j => PX P i (PaddI (Pdouble_minus_one j) Q')
     | xI j => PX P i (PaddI (xO j) Q')
     end 
   end.

  Fixpoint PsubI (j:positive) (P:Pol){struct P} : Pol :=
   match P with
   | Pc c => mkPinj j (PaddC (--Q) c)
   | Pinj j' Q' => 
     match ZPminus j' j with
     | Zpos k =>  mkPinj j (Pop (Pinj k Q') Q)
     | Z0 => mkPinj j (Pop Q' Q)
     | Zneg k => mkPinj j' (PsubI k Q')
     end
   | PX P i Q' => 
     match j with
     | xH => PX P i (Pop Q' Q)
     | xO j => PX P i (PsubI (Pdouble_minus_one j) Q')
     | xI j => PX P i (PsubI (xO j) Q')
     end 
   end.
 
 Variable P' : Pol.
 
 Fixpoint PaddX (i':positive) (P:Pol) {struct P} : Pol :=
  match P with
  | Pc c => PX P' i' P
  | Pinj j Q' =>
    match j with
    | xH =>  PX P' i' Q'
    | xO j => PX P' i' (Pinj (Pdouble_minus_one j) Q')
    | xI j => PX P' i' (Pinj (xO j) Q')
    end
  | PX P i Q' =>
    match ZPminus i i' with
    | Zpos k => mkPX (Pop (PX P k P0) P') i' Q'
    | Z0 => mkPX (Pop P P') i Q'
    | Zneg k => mkPX (PaddX k P) i Q'
    end
  end.

 Fixpoint PsubX (i':positive) (P:Pol) {struct P} : Pol :=
  match P with
  | Pc c => PX (--P') i' P
  | Pinj j Q' =>
    match j with
    | xH =>  PX (--P') i' Q'
    | xO j => PX (--P') i' (Pinj (Pdouble_minus_one j) Q')
    | xI j => PX (--P') i' (Pinj (xO j) Q')
    end
  | PX P i Q' =>
    match ZPminus i i' with
    | Zpos k => mkPX (Pop (PX P k P0) P') i' Q'
    | Z0 => mkPX (Pop P P') i Q'
    | Zneg k => mkPX (PsubX k P) i Q'
    end
  end.

    
 End PopI.

 Fixpoint Padd (P P': Pol) {struct P'} : Pol :=
  match P' with
  | Pc c' => PaddC P c'
  | Pinj j' Q' => PaddI Padd Q' j' P
  | PX P' i' Q' =>
    match P with
    | Pc c => PX P' i' (PaddC Q' c)
    | Pinj j Q =>  
      match j with
      | xH => PX P' i' (Padd Q Q')
      | xO j => PX P' i' (Padd (Pinj (Pdouble_minus_one j) Q) Q')
      | xI j => PX P' i' (Padd (Pinj (xO j) Q) Q')
      end 
    | PX P i Q =>
      match ZPminus i i' with
      | Zpos k => mkPX (Padd (PX P k P0) P') i' (Padd Q Q')
      | Z0 => mkPX (Padd P P') i (Padd Q Q')
      | Zneg k => mkPX (PaddX Padd P' k P) i (Padd Q Q')
      end
    end
  end.
 Notation "P ++ P'" := (Padd P P').

 Fixpoint Psub (P P': Pol) {struct P'} : Pol :=
  match P' with
  | Pc c' => PsubC P c'
  | Pinj j' Q' => PsubI Psub Q' j' P
  | PX P' i' Q' =>
    match P with
    | Pc c => PX (--P') i' (*(--(PsubC Q' c))*) (PaddC (--Q') c)
    | Pinj j Q =>  
      match j with
      | xH => PX (--P') i' (Psub Q Q')
      | xO j => PX (--P') i' (Psub (Pinj (Pdouble_minus_one j) Q) Q')
      | xI j => PX (--P') i' (Psub (Pinj (xO j) Q) Q')
      end 
    | PX P i Q =>
      match ZPminus i i' with
      | Zpos k => mkPX (Psub (PX P k P0) P') i' (Psub Q Q')
      | Z0 => mkPX (Psub P P') i (Psub Q Q')
      | Zneg k => mkPX (PsubX Psub P' k P) i (Psub Q Q')
      end
    end
  end.
 Notation "P -- P'" := (Psub P P').
 
 (** Multipilcation *) 

 Fixpoint PmulC_aux (P:Pol) (c:C) {struct P} : Pol :=
  match P with
  | Pc c' => Pc (c' *! c)
  | Pinj j Q => mkPinj j (PmulC_aux Q c)
  | PX P i Q => mkPX (PmulC_aux P c) i (PmulC_aux Q c)
  end.

 Definition PmulC P c :=
  if c ?=! cO then P0 else
  if c ?=! cI then P else PmulC_aux P c.
 
 Section PmulI. 
  Variable Pmul : Pol -> Pol -> Pol.
  Variable Q : Pol.
  Fixpoint PmulI (j:positive) (P:Pol) {struct P} : Pol :=
   match P with
   | Pc c => mkPinj j (PmulC Q c)
   | Pinj j' Q' => 
     match ZPminus j' j with
     | Zpos k => mkPinj j (Pmul (Pinj k Q') Q)
     | Z0 => mkPinj j (Pmul Q' Q)
     | Zneg k => mkPinj j' (PmulI k Q')
     end
   | PX P' i' Q' =>
     match j with
     | xH => mkPX (PmulI xH P') i' (Pmul Q' Q)
     | xO j' => mkPX (PmulI j P') i' (PmulI (Pdouble_minus_one j') Q')
     | xI j' => mkPX (PmulI j P') i' (PmulI (xO j') Q')
     end
   end.
 
 End PmulI.

 Fixpoint Pmul_aux (P P' : Pol) {struct P'} : Pol :=
  match P' with
  | Pc c' => PmulC P c'
  | Pinj j' Q' => PmulI Pmul_aux Q' j' P
  | PX P' i' Q' => 
     (mkPX (Pmul_aux P P') i' P0) ++ (PmulI Pmul_aux Q' xH P)
  end.

 Definition Pmul P P' :=
  match P with
  | Pc c => PmulC P' c
  | Pinj j Q => PmulI Pmul_aux Q j P'
  | PX P i Q => 
    Padd (mkPX (Pmul_aux P P') i P0) (PmulI Pmul_aux Q xH P')
  end.
 Notation "P ** P'" := (Pmul P P').
 
 (** Evaluation d'un polynome vers R *)

 Fixpoint pow (x:R) (i:positive) {struct i}: R :=
  match i with
  | xH => x
  | xO i => let p := pow x i in p * p
  | xI i => let p := pow x i in x * p * p
  end.

 Fixpoint Pphi(l:list R) (P:Pol) {struct P} : R :=
  match P with
  | Pc c => [c]
  | Pinj j Q => Pphi (jump j l) Q
  | PX P i Q => 
     let x := hd 0 l in
     let xi := pow x i in
    (Pphi l P) * xi + (Pphi (tl l) Q)      
  end.

 Reserved Notation "P @ l " (at level 10, no associativity).
 Notation "P @ l " := (Pphi l P).
 (** Preuves *)
 Lemma ZPminus_spec : forall x y,
  match ZPminus x y with
  | Z0 => x = y
  | Zpos k => x = (y + k)%positive
  | Zneg k => y = (x + k)%positive
  end.
 Proof.
  induction x;destruct y.
  replace (ZPminus (xI x) (xI y)) with (Zdouble (ZPminus x y));trivial.
  assert (H := IHx y);destruct (ZPminus x y);unfold Zdouble;rewrite H;trivial.
  replace (ZPminus (xI x) (xO y)) with (Zdouble_plus_one (ZPminus x y));trivial.
  assert (H := IHx y);destruct (ZPminus x y);unfold Zdouble_plus_one;rewrite H;trivial.
  apply Pplus_xI_double_minus_one.
  simpl;trivial.
  replace (ZPminus (xO x) (xI y)) with (Zdouble_minus_one (ZPminus x y));trivial.
  assert (H := IHx y);destruct (ZPminus x y);unfold Zdouble_minus_one;rewrite H;trivial.
  apply Pplus_xI_double_minus_one.
  replace (ZPminus (xO x) (xO y)) with (Zdouble (ZPminus x y));trivial.
  assert (H := IHx y);destruct (ZPminus x y);unfold Zdouble;rewrite H;trivial.
  replace (ZPminus (xO x) xH) with (Zpos (Pdouble_minus_one x));trivial.
  rewrite <- Pplus_one_succ_l.
  rewrite Psucc_o_double_minus_one_eq_xO;trivial.
  replace (ZPminus xH (xI y)) with (Zneg (xO y));trivial.
  replace (ZPminus xH (xO y)) with (Zneg (Pdouble_minus_one y));trivial.
  rewrite <- Pplus_one_succ_l.
  rewrite Psucc_o_double_minus_one_eq_xO;trivial.
  simpl;trivial.
 Qed.

 Lemma pow_Psucc : forall x j, pow x (Psucc j) == x * pow x j.
 Proof.
  induction j;simpl;rsimpl.
  rewrite IHj;rsimpl;mul_push x;rrefl.
 Qed.

 Lemma pow_Pplus : forall x i j, pow x (i + j) == pow x i * pow x j.
 Proof.
  intro x;induction i;intros.
  rewrite xI_succ_xO;rewrite Pplus_one_succ_r.
  rewrite <- Pplus_diag;repeat rewrite <- Pplus_assoc.
  rewrite (IHi (Pplus i (Pplus xH j)));rewrite (IHi (Pplus xH j)).
  rewrite (IHi (Pplus i xH));rewrite (IHi xH). 
  rewrite Pplus_comm;rewrite <- Pplus_one_succ_r;rewrite (pow_Psucc x j).
  simpl;rsimpl.
  rewrite <- Pplus_diag;repeat rewrite <- Pplus_assoc.
  rewrite (IHi (Pplus i j));rewrite (IHi j);rewrite (IHi i);rsimpl.
  rewrite Pplus_comm;rewrite <- Pplus_one_succ_r;rewrite (pow_Psucc x j);
   simpl;rsimpl.
 Qed.
  
 Lemma Peq_ok : forall P P', 
    (P ?== P') = true -> forall l, P@l == P'@ l.
 Proof.
  induction P;destruct P';simpl;intros;try discriminate;trivial.
  apply CRmorph.(morph_eq);trivial.
  assert (H1 := Pcompare_Eq_eq p p0); destruct ((p ?= p0)%positive Eq);
   try discriminate H.
  rewrite (IHP P' H (jump p l)); rewrite H1;trivial;rrefl.
  assert (H1 := Pcompare_Eq_eq p p0); destruct ((p ?= p0)%positive Eq);
   try discriminate H.
  rewrite H1;trivial. clear H1.
  assert (H1 := IHP1 P'1);assert (H2 := IHP2 P'2);
   destruct (P2 ?== P'1);[destruct (P3 ?== P'2); [idtac|discriminate H]
                         |discriminate H].
  rewrite (H1 H l);rewrite (H2 H (tl l));rrefl.
 Qed.

 Lemma Pphi0 : forall l, P0@l == 0.
 Proof.
  intros;simpl;apply CRmorph.(morph0).
 Qed.

 Lemma Pphi1 : forall l,  P1@l == 1.
 Proof.
  intros;simpl;apply CRmorph.(morph1).
 Qed.

 Lemma mkPinj_ok : forall j l P, (mkPinj j P)@l == P@(jump j l).
 Proof.
  intros j l p;destruct p;simpl;rsimpl.
  rewrite <-jump_Pplus;rewrite Pplus_comm;rrefl.
 Qed.

 Lemma mkPX_ok : forall l P i Q, 
  (mkPX P i Q)@l == P@l*(pow (hd 0 l) i) + Q@(tl l).
 Proof.
  intros l P i Q;unfold mkPX.
  destruct P;try (simpl;rrefl).
  assert (H := @CRmorph.(morph_eq) c cO);destruct (c ?=! cO);simpl;try rrefl.
  rewrite (H (refl_equal true));rewrite CRmorph.(morph0).
  rewrite (mkPinj_ok xH l Q);rsimpl;simpl;rrefl.
  assert (H := @Peq_ok P3 P0);destruct (P3 ?== P0);simpl;try rrefl.
  rewrite (H (refl_equal true) (tl l));trivial.
  rewrite (Pphi0 (tl l));rewrite (pow_Pplus (hd 0 l) p i);rsimpl.
 Qed.

 Ltac Esimpl :=
  repeat (progress (
   match goal with
   | |- context [P0@?l] => rewrite (Pphi0 l)
   | |- context [P1@?l] => rewrite (Pphi1 l)
   | |- context [(mkPinj ?j ?P)@?l] => rewrite (mkPinj_ok j l P)
   | |- context [(mkPX ?P ?i ?Q)@?l] => rewrite (mkPX_ok l P i Q)
   | |- context [[cO]] => rewrite CRmorph.(morph0)
   | |- context [[cI]] => rewrite CRmorph.(morph1)
   | |- context [[?x +! ?y]] => rewrite (CRmorph.(morph_add) x y)
   | |- context [[?x *! ?y]] => rewrite (CRmorph.(morph_mul) x y)
   | |- context [[?x -! ?y]] => rewrite (CRmorph.(morph_sub) x y)
   | |- context [[-! ?x]] => rewrite (CRmorph.(morph_opp) x)
   end));
  rsimpl; simpl. 
 
 Lemma PaddC_ok : forall c P l, (PaddC P c)@l == P@l + [c].
 Proof.
  induction P;simpl;intros;Esimpl;trivial.
  rewrite (IHP2 (tl l));rsimpl.
 Qed.

 Lemma PsubC_ok : forall c P l, (PsubC P c)@l == P@l - [c].
 Proof.
  induction P;simpl;intros.
  Esimpl.
  rewrite (IHP (jump p l));rsimpl.
  rewrite (IHP2 (tl l));rsimpl.
 Qed.

 Lemma PmulC_aux_ok : forall c P l, (PmulC_aux P c)@l == P@l * [c].
 Proof.
  induction P;simpl;intros;Esimpl;trivial.
  rewrite (IHP1 l);rewrite (IHP2 (tl l));rsimpl.
  mul_push ([c]);rrefl.
 Qed. 

 Lemma PmulC_ok : forall c P l, (PmulC P c)@l == P@l * [c].
 Proof.
  intros c P l; unfold PmulC.
  assert (H:= @CRmorph.(morph_eq) c cO);destruct (c ?=! cO).
  rewrite (H (refl_equal true));Esimpl.
  assert (H1:= @CRmorph.(morph_eq) c cI);destruct (c ?=! cI).
  rewrite (H1 (refl_equal true));Esimpl.
  apply PmulC_aux_ok.
 Qed.

 Lemma Popp_ok : forall P l, (--P)@l == - P@l.
 Proof.
  induction P;simpl;intros.
  Esimpl.
  apply IHP.
  rewrite (IHP1 l);rewrite (IHP2 (tl l));rsimpl.
 Qed.

 Ltac Esimpl2 :=
  Esimpl;
  repeat (progress (
   match goal with 
   | |- context [(PaddC ?P ?c)@?l] => rewrite (PaddC_ok c P l)
   | |- context [(PsubC ?P ?c)@?l] => rewrite (PsubC_ok c P l)
   | |- context [(PmulC ?P ?c)@?l] => rewrite (PmulC_ok c P l)
   | |- context [(--?P)@?l] => rewrite (Popp_ok P l)
   end)); Esimpl.

 Lemma Padd_ok : forall P' P l, (P ++ P')@l == P@l + P'@l.
 Proof.
  induction P';simpl;intros;Esimpl2.
  generalize P p l;clear P p l.
  induction P;simpl;intros.
  Esimpl2;apply ARth.(ARadd_sym).
  assert (H := ZPminus_spec p p0);destruct (ZPminus p p0).
  rewrite H;Esimpl. rewrite (IHP' P (jump p0 l));rrefl.
  rewrite H;Esimpl. rewrite (IHP' (Pinj p1 P) (jump p0 l));Esimpl.
  rewrite <- jump_Pplus;rewrite Pplus_comm;rrefl.
  rewrite H;Esimpl. rewrite (IHP p1 (jump p l)).
  rewrite <- jump_Pplus;rewrite Pplus_comm;rrefl.
  destruct p0;simpl.
  rewrite (IHP2 (xO p0) (tl l));simpl;rsimpl.
  rewrite (IHP2 (Pdouble_minus_one p0) (tl l));simpl.
  rewrite jump_Pdouble_minus_one;rsimpl.
  rewrite (IHP' P3 (tl l));rsimpl.
  destruct P;simpl. 
  Esimpl2;add_push [c];rrefl.
  destruct p0;simpl;Esimpl2.
  rewrite (IHP'2 (Pinj (xO p0) P) (tl l));simpl.
  rsimpl;add_push (P'1@l * (pow (hd 0 l) p));rrefl.
  rewrite (IHP'2 (Pinj (Pdouble_minus_one p0) P) (tl l));simpl.
  rewrite jump_Pdouble_minus_one;rsimpl;add_push (P'1@l * (pow (hd 0 l) p));rrefl.
  rewrite (IHP'2 P (tl l));rsimpl. add_push (P @ (tl l));rrefl.
  assert (H := ZPminus_spec p0 p);destruct (ZPminus p0 p);Esimpl2.
  rewrite (IHP'1 P2 l);rewrite (IHP'2 P3 (tl l));rsimpl.
  add_push (P3 @ (tl l));rewrite H;rrefl.
  rewrite (IHP'1 (PX P2 p1 P0) l);rewrite (IHP'2 P3 (tl l));simpl;Esimpl.
  rewrite H;rewrite Pplus_comm.
  rewrite (pow_Pplus (hd 0 l) p1 p);rsimpl.
  add_push (P3 @ (tl l));rrefl.
  assert (forall P k l, 
           (PaddX Padd P'1 k P) @ l == P@l + P'1@l * pow (hd 0 l) k).
   induction P;simpl;intros;try apply ARth.(ARadd_sym).
   destruct p2;simpl;try apply ARth.(ARadd_sym).
   rewrite jump_Pdouble_minus_one;apply ARth.(ARadd_sym).
    assert (H1 := ZPminus_spec p2 k);destruct (ZPminus p2 k);Esimpl2.
    rewrite (IHP'1 P4 l0);rsimpl.    rewrite H1;add_push (P5 @ (tl l0));rrefl.
    rewrite (IHP'1 (PX P4 p3 P0) l0);simpl;Esimpl.
    rewrite H1;rewrite Pplus_comm.
    rewrite (pow_Pplus (hd 0 l0) p3 k);simpl;Esimpl.
    add_push (P5 @ (tl l0));rrefl.
    rewrite (IHP1 p3 l0);rewrite H1;rewrite Pplus_comm.
    rewrite (pow_Pplus (hd 0 l0) p3 p2);simpl;rsimpl.
    add_push (P5 @ (tl l0));rrefl.
  rewrite (H0 P2 p1 l);rsimpl.
  add_push (P3 @ (tl l)).
  rewrite H;rewrite Pplus_comm.
  rewrite (IHP'2 P3 (tl l));rewrite (pow_Pplus (hd 0 l) p1 p0);rsimpl.
  add_push (P3 @ (tl l));rrefl.
 Qed.

 Lemma Psub_ok : forall P' P l, (P -- P')@l == P@l - P'@l.
 Proof.
  induction P';simpl;intros;Esimpl2;trivial.
  generalize P p l;clear P p l.
  induction P;simpl;intros.
  Esimpl2;apply ARth.(ARadd_sym).
  assert (H := ZPminus_spec p p0);destruct (ZPminus p p0).
  rewrite H;Esimpl.  rewrite (IHP' P (jump p0 l));rsimpl. 
  rewrite H;Esimpl. rewrite (IHP' (Pinj p1 P) (jump p0 l));Esimpl.
  rewrite <- jump_Pplus;rewrite Pplus_comm;rrefl.
  rewrite H;Esimpl. rewrite (IHP p1 (jump p l)).
  rewrite <- jump_Pplus;rewrite Pplus_comm;rrefl.
  destruct p0;simpl.
  rewrite (IHP2 (xO p0) (tl l));simpl;rsimpl.
  rewrite (IHP2 (Pdouble_minus_one p0) (tl l));simpl.
  rewrite jump_Pdouble_minus_one;rsimpl.
  rewrite (IHP' P3 (tl l));rsimpl. 
  destruct P;simpl. 
  repeat rewrite Popp_ok;Esimpl2;rsimpl;add_push [c];try rrefl.
  destruct p0;simpl;Esimpl2.
  rewrite (IHP'2 (Pinj (xO p0) P) (tl l));simpl;rsimpl;add_push (P'1@l * (pow (hd 0 l) p));trivial.
  add_push (P @ (jump p0 (jump p0 (tl l))));rrefl.
  rewrite (IHP'2 (Pinj (Pdouble_minus_one p0) P) (tl l));simpl;rewrite jump_Pdouble_minus_one;rsimpl.
  add_push (- (P'1 @ l * pow (hd 0 l) p));rrefl.
  rewrite (IHP'2 P (tl l));rsimpl;add_push (P @ (tl l));rrefl.
  assert (H := ZPminus_spec p0 p);destruct (ZPminus p0 p);Esimpl2.
  rewrite (IHP'1 P2 l); rewrite (IHP'2 P3 (tl l));rsimpl.
  add_push (P3 @ (tl l));rewrite H;rrefl.
   rewrite (IHP'1 (PX P2 p1 P0) l); rewrite (IHP'2 P3 (tl l));rsimpl;simpl;Esimpl.
  rewrite H;rewrite Pplus_comm.
  rewrite (pow_Pplus (hd 0 l) p1 p);rsimpl.
  add_push (P3 @ (tl l));rrefl.
  assert (forall P k l, 
           (PsubX Psub P'1 k P) @ l == P@l + - P'1@l * pow (hd 0 l) k).
   induction P;simpl;intros.
   rewrite (Popp_ok P'1 l0);rsimpl;apply ARth.(ARadd_sym);trivial.
   destruct p2;simpl;rewrite (Popp_ok P'1 l0);rsimpl.
   apply ARth.(ARadd_sym);trivial.
   rewrite jump_Pdouble_minus_one;apply ARth.(ARadd_sym);trivial.
   apply ARth.(ARadd_sym);trivial.
    assert (H1 := ZPminus_spec p2 k);destruct (ZPminus p2 k);Esimpl2;rsimpl.
    rewrite (IHP'1 P4 l0);rsimpl;add_push (P5 @ (tl l0));rewrite H1;rrefl.
    rewrite (IHP'1 (PX P4 p3 P0) l0);rewrite H1;rewrite Pplus_comm.
    rewrite (pow_Pplus (hd 0 l0) p3 k);simpl;Esimpl.
    add_push (P5 @ (tl l0));rrefl.
    rewrite (IHP1 p3 l0);rewrite H1;rewrite Pplus_comm.
    rewrite (pow_Pplus (hd 0 l0) p3 p2);simpl;rsimpl.
    add_push (P5 @ (tl l0));rrefl.
  rewrite (H0 P2 p1 l);rsimpl.
  rewrite (IHP'2 P3 (tl l));rsimpl;add_push (P3 @ (tl l)).
  rewrite H;rewrite Pplus_comm.
  rewrite (pow_Pplus (hd 0 l) p1 p0);rsimpl.
 Qed.
 
 Lemma PmulI_ok : 
  forall P', 
   (forall (P : Pol) (l : list R), (Pmul_aux P P') @ l == P @ l * P' @ l) ->
   forall (P : Pol) (p : positive) (l : list R),
    (PmulI Pmul_aux P' p P) @ l == P @ l * P' @ (jump p l).
 Proof.
  induction P;simpl;intros.
  Esimpl2;apply ARth.(ARmul_sym).
  assert (H1 := ZPminus_spec p p0);destruct (ZPminus p p0);Esimpl2.
  rewrite H1; rewrite (H P (jump p0 l));rrefl.
  rewrite H1;rewrite (H (Pinj p1 P) (jump p0 l)).
  rewrite Pplus_comm.
  rewrite (jump_Pplus p1 p0 l);simpl;rrefl.
  rewrite H1;rewrite Pplus_comm.
  rewrite (jump_Pplus p1 p l);rewrite (IHP p1 (jump p l));rrefl.
  destruct p0;Esimpl2.
  rewrite (IHP1 (xI p0) l);rewrite (IHP2 (xO p0) (tl l));simpl;rsimpl.
  mul_push (pow (hd 0 l) p);rrefl.
  rewrite (IHP1 (xO p0) l);rewrite (IHP2 (Pdouble_minus_one p0) (tl l));simpl;rsimpl.
  mul_push (pow (hd 0 l) p); rewrite jump_Pdouble_minus_one;rrefl.
  rewrite (IHP1 xH l);simpl;rsimpl.
  mul_push (pow (hd 0 l) p).
  rewrite (H P3 (tl l));rrefl.
 Qed.

 Lemma Pmul_aux_ok : forall P' P l,(Pmul_aux P P')@l == P@l * P'@l.
 Proof.
  induction P';simpl;intros.
  Esimpl2;trivial.
  apply PmulI_ok;trivial.
  rewrite (Padd_ok (PmulI Pmul_aux P'2 xH P) (mkPX (Pmul_aux P P'1) p P0) l);Esimpl2.
  rewrite (PmulI_ok P'2 IHP'2 P xH l). rewrite (IHP'1 P l). rrefl.
 Qed.

 Lemma Pmul_ok : forall P P' l, (P**P')@l == P@l * P'@l.
 Proof.
  destruct P;simpl;intros.
  Esimpl2;apply ARth.(ARmul_sym).
  rewrite (PmulI_ok P (Pmul_aux_ok P) P' p l).
  apply ARth.(ARmul_sym).
  rewrite (Padd_ok (PmulI Pmul_aux P3 xH P') (mkPX (Pmul_aux P2 P') p P0) l);
   Esimpl2.
  rewrite (PmulI_ok P3 (Pmul_aux_ok P3) P' xH l);trivial.
  rewrite (Pmul_aux_ok P' P2 l);mul_push (P' @ l).
  rewrite (ARmul_sym ARth (P' @ l) (P3 @ (jump xH l)));rrefl.
 Qed.

 (** Definition des expressions polynomiales *)

 Inductive PExpr : Type :=
  | PEc : C -> PExpr
  | PEX : positive -> PExpr
  | PEadd : PExpr -> PExpr -> PExpr
  | PEsub : PExpr -> PExpr -> PExpr
  | PEmul : PExpr -> PExpr -> PExpr
  | PEopp : PExpr -> PExpr.

 (** normalisation vers les polynomes *)
 
 Definition X := (PX P1 xH P0).

 Definition mkX j :=
  match j with
  | xH => X
  | xO j => Pinj (Pdouble_minus_one j) X
  | xI j => Pinj (xO j) X
  end.

 Fixpoint norm (pe:PExpr) : Pol :=
  match pe with
  | PEc c => Pc c
  | PEX j => mkX j
  | PEadd pe1 (PEopp pe2) => Psub (norm pe1) (norm pe2)
  | PEadd (PEopp pe1) pe2 => Psub (norm pe2) (norm pe1)
  | PEadd pe1 pe2 => Padd (norm pe1) (norm pe2)
  | PEsub pe1 pe2 => Psub (norm pe1) (norm pe2)
  | PEmul pe1 pe2 => Pmul (norm pe1) (norm pe2)
  | PEopp pe1 => Popp (norm pe1)
  end.

   (* syntaxification *)
 Ltac mkPolexpr Cst add mul sub opp t fv := 
 let rec mkP t :=
  match t with 
  | (add ?t1 ?t2) => 
    let e1 := mkP t1 in
    let e2 := mkP t2 in constr:(PEadd e1 e2)
  | (mul ?t1 ?t2) => 
    let e1 := mkP t1 in
    let e2 := mkP t2 in constr:(PEmul e1 e2)
  | (sub ?t1 ?t2) => 
    let e1 := mkP t1 in
    let e2 := mkP t2 in constr:(PEsub e1 e2)
  | (opp ?t1) =>
    let e1 := mkP t1 in constr:(PEopp e1)
  | _ =>
    match Cst t with
    | RBfalse => let p := Find_at t fv in constr:(PEX p)
    | ?c => constr:(PEc c)
    end
  end
 in mkP t.

 (** evaluation des expressions polynomiales dans R *)
 
 Fixpoint PEeval (l:list R) (pe:PExpr) {struct pe} : R :=
  match pe with
  | PEc c => phi c
  | PEX j => nth 0 j l
  | PEadd pe1 pe2 => (PEeval l pe1) + (PEeval l pe2)
  | PEsub pe1 pe2 => (PEeval l pe1) - (PEeval l pe2)
  | PEmul pe1 pe2 => (PEeval l pe1) * (PEeval l pe2)
  | PEopp pe1 => - (PEeval l pe1)
  end.

 (** Preuves de correction *)
 

 Lemma mkX_ok : forall p l, nth 0 p l == (mkX p) @ l.
 Proof.
  destruct p;simpl;intros;Esimpl;trivial.
  rewrite <-jump_tl;rewrite nth_jump;rrefl.
  rewrite <- nth_jump.
  rewrite nth_Pdouble_minus_one;rrefl.
 Qed.

 Lemma norm_PEopp : forall l pe,  (norm (PEopp pe))@l == -(norm pe)@l.
 Proof.
  intros;simpl;apply Popp_ok.
 Qed.

 Ltac Esimpl3 := 
  repeat match goal with
  | |- context [(?P1 ++ ?P2)@?l] => rewrite (Padd_ok P2 P1 l)
  | |- context [(?P1 -- ?P2)@?l] => rewrite (Psub_ok P2 P1 l)
  | |- context [(norm (PEopp ?pe))@?l] => rewrite (norm_PEopp l pe)
  end;Esimpl2;try rrefl;try apply ARth.(ARadd_sym).

 Lemma norm_ok : forall l pe,  PEeval l pe == (norm pe)@l.
 Proof.
  induction pe;simpl;Esimpl3.
  apply mkX_ok.
  destruct pe1;destruct pe2;rewrite IHpe1;rewrite IHpe2;Esimpl3.
  rewrite IHpe1;rewrite IHpe2;rrefl.
  rewrite (Pmul_ok (norm pe1) (norm pe2) l);rewrite IHpe1;rewrite IHpe2;rrefl.
  rewrite IHpe;rrefl.
 Qed.

 Lemma ring_correct : forall l pe1 pe2, 
  ((norm pe1) ?== (norm pe2)) = true -> (PEeval l pe1) == (PEeval l pe2).
 Proof.
  intros l pe1 pe2 H.
  rewrite (norm_ok l pe1);rewrite (norm_ok l pe2).
  apply  (Peq_ok (norm pe1) (norm pe2) H l).
 Qed.

(** Fonction d'evaluation sans parantheses *)
 Fixpoint mkmult (r:R) (lm:list R) {struct lm}: R :=
  match lm with
  | nil => r
  | cons h t => mkmult (r*h) t
  end.
 		       
 Definition mkadd_mult rP lm :=
  match lm with
  | nil => rP + 1
  | cons h t => rP + mkmult h t
  end.

 Fixpoint powl (i:positive) (x:R) (l:list R) {struct i}: list R :=
  match i with
  | xH => cons x l
  | xO i => powl i x (powl i x l)
  | xI i => powl i x (powl i x (cons x l))
  end.
			       
 Fixpoint add_mult_dev (rP:R) (P:Pol) (fv lm:list R) {struct P} : R :=
  (* rP + P@l * lm *)
  match P with
  | Pc c => if c ?=! cI then mkadd_mult rP (rev lm) 
    else mkadd_mult rP (cons [c] (rev lm))
  | Pinj j Q => add_mult_dev rP Q (jump j fv) lm
  | PX P i Q => 
     let rP := add_mult_dev rP P fv (powl i (hd 0 fv) lm) in
     if Q ?== P0 then rP else add_mult_dev rP Q (tl fv) lm
  end.
 
 Definition mkmult1 lm :=
  match lm with
  | nil => rI
  | cons h t => mkmult h t
  end.
    
 Fixpoint mult_dev (P:Pol) (fv lm : list R) {struct P} : R :=
  (* P@l * lm *)							      
  match P with
  | Pc c => if c ?=! cI then mkmult1 (rev lm) else mkmult [c] (rev lm)
  | Pinj j Q => mult_dev Q (jump j fv) lm
  | PX P i Q => 
     let rP := mult_dev P fv (powl i (hd 0 fv) lm) in
     if Q ?== P0 then rP else add_mult_dev rP Q (tl fv) lm
  end. 

 Definition Pphi_dev fv P := mult_dev P fv (nil R).

 Add Morphism mkmult : mkmult_ext.
  intros r r0 l;generalize l r r0;clear l r r0;induction l;simpl;intros.
  trivial. apply IHl; rewrite H;rrefl.
 Qed.

 Lemma mul_mkmult : forall lm r1 r2, r1 * mkmult r2 lm == mkmult (r1*r2) lm.
 Proof.
  induction lm;simpl;intros;try rrefl.
  rewrite (IHlm r1 (r2 * a)).
  setoid_replace (r1 * (r2 * a)) with (r1 * r2 * a);Esimpl.
 Qed.

 Lemma mkmult1_mkmult : forall lm r, r * mkmult1 lm == mkmult r lm.
 Proof.
  destruct lm;simpl;intros. Esimpl.
  apply mul_mkmult.
 Qed.
 
 Lemma mkmult1_mkmult_1 : forall lm, mkmult1 lm == mkmult 1 lm.
 Proof.
  intros;rewrite <- (mkmult1_mkmult lm 1);Esimpl.
 Qed.

 Lemma mkmult_rev_append : forall lm l r,
  mkmult r (rev_append l lm) == mkmult (mkmult r l) lm.
 Proof.
  induction lm;simpl;intros. rrefl.
  rewrite (IHlm (cons a l) r);simpl.
  rewrite (ARth.(ARmul_sym) r a); rewrite <- (mul_mkmult l a r).
   rewrite <- (mul_mkmult lm a (mkmult r l)).
  rewrite (ARth.(ARmul_sym) (mkmult r l) a); 
   rewrite <- (mul_mkmult lm a (mkmult r l));rrefl.
 Qed.

 Lemma powl_mkmult_rev : forall p r x lm,
  mkmult r (rev (powl p x lm)) == mkmult (pow x p * r) (rev lm).
 Proof.
  induction p;simpl;intros.
  rewrite (IHp r x (powl p x (cons x lm))).
  rewrite (IHp (pow x p * r) x (cons x lm)).
  unfold rev;simpl.
  rewrite (mkmult_rev_append lm (cons x (nil R)) (pow x p * (pow x p * r))).
  simpl.
  setoid_replace (pow x p * (pow x p * r) * x) 
    with (x * pow x p * pow x p * r);Esimpl.
  mul_push x;rrefl.
  rewrite (mkmult_rev_append lm (nil R) (x * pow x p * pow x p * r));
   simpl;rrefl.
  rewrite (IHp r x (powl p x lm)).
  rewrite (IHp (pow x p * r) x lm).
  setoid_replace (pow x p * (pow x p * r) ) 
    with (pow x p * pow x p * r);Esimpl.
  unfold rev;simpl. rewrite (mkmult_rev_append lm (cons x (nil R)) r);simpl.
  rewrite (mkmult_rev_append lm (nil R) (x*r));simpl.
  rewrite (ARth.(ARmul_sym) r x);rrefl.
 Qed.

 Lemma Pphi_add_mult_dev : forall P rP fv lm, 
    rP + P@fv * mkmult1 (rev lm) == add_mult_dev rP P fv lm.
 Proof.
  induction P;simpl;intros.
  assert (H := CRmorph.(morph_eq) c cI).
   destruct (c ?=! cI).
    rewrite (H (refl_equal true));rewrite CRmorph.(morph1);Esimpl.
    destruct (rev lm);Esimpl;rrefl.
    rewrite (mkmult1_mkmult (rev lm) [c]);rrefl.
   apply IHP.
   replace (match P3 with
       | Pc c => c ?=! cO
       | Pinj _ _ => false
       | PX _ _ _ => false
       end) with (Peq P3 P0);trivial.
   assert (H := Peq_ok P3 P0).
    destruct (P3 ?== P0).
    rewrite (H (refl_equal true) (tl fv));simpl;Esimpl.
   rewrite <- (IHP1 rP fv (powl p (hd 0 fv) lm)).
     rewrite (mkmult1_mkmult_1 (rev (powl p (hd 0 fv) lm))).
     rewrite (mkmult1_mkmult_1 (rev lm)). 
     rewrite (powl_mkmult_rev p 1 (hd 0 fv) lm).
     rewrite <- (mul_mkmult (rev lm) (pow (hd 0 fv) p) 1);Esimpl.
     rewrite <-(IHP2 (add_mult_dev rP P2 fv (powl p (hd 0 fv) lm)) (tl fv) lm).
   rewrite <- (IHP1 rP fv (powl p (hd 0 fv) lm)).
     rewrite (mkmult1_mkmult_1 (rev (powl p (hd 0 fv) lm))).
     rewrite (powl_mkmult_rev p 1 (hd 0 fv) lm).
     rewrite <- (mul_mkmult (rev lm) (pow (hd 0 fv) p) 1);Esimpl.
     rewrite <- (mkmult1_mkmult_1 (rev lm));rrefl.
 Qed.
                                     
 Lemma Pphi_mult_dev : forall P fv lm,
	 P@fv * mkmult1 (rev lm) == mult_dev P fv lm.
 Proof.
  induction P;simpl;intros.
   assert (H := CRmorph.(morph_eq) c cI).
   destruct (c ?=! cI).
    rewrite (H (refl_equal true));rewrite CRmorph.(morph1);Esimpl.
    apply  mkmult1_mkmult.
   apply IHP.
   replace (match P3 with
       | Pc c => c ?=! cO
       | Pinj _ _ => false
       | PX _ _ _ => false
       end) with (Peq P3 P0);trivial.
   assert (H := Peq_ok P3 P0).
   destruct (P3 ?== P0).
   rewrite (H (refl_equal true) (tl fv));simpl;Esimpl.
   rewrite <- (IHP1 fv (powl p (hd 0 fv) lm)).
   rewrite (mkmult1_mkmult_1 (rev (powl p (hd 0 fv) lm))).
   rewrite (mkmult1_mkmult_1 (rev lm)). 
   rewrite (powl_mkmult_rev p 1 (hd 0 fv) lm).
   rewrite <- (mul_mkmult (rev lm) (pow (hd 0 fv) p) 1);Esimpl.
   rewrite <- (Pphi_add_mult_dev P3 (mult_dev P2 fv (powl p (hd 0 fv) lm)) 
                  (tl fv) lm).
   rewrite <- (IHP1 fv (powl p (hd 0 fv) lm)).
   rewrite (mkmult1_mkmult_1 (rev (powl p (hd 0 fv) lm))).
   rewrite (powl_mkmult_rev p 1 (hd 0 fv) lm).
   rewrite <- (mul_mkmult (rev lm) (pow (hd 0 fv) p) 1);Esimpl.
   rewrite <- (mkmult1_mkmult_1 (rev lm));rrefl.
 Qed.

 Lemma Pphi_Pphi_dev :  forall P l,  P@l == Pphi_dev l P.
 Proof. 		
  unfold Pphi_dev;intros.
  rewrite <- (Pphi_mult_dev P l (nil R));simpl;Esimpl.
 Qed.

 Lemma Pphi_dev_ok :  forall l pe,  PEeval l pe == Pphi_dev l (norm pe).
 Proof.
  intros l pe;rewrite <- (Pphi_Pphi_dev (norm pe) l);apply norm_ok.
 Qed.
 
 (** La tactic qui fait la meme chose *)
 Ltac Trev l :=  
  let rec rev_append rev l :=
   match l with
   | (nil _) => constr:(rev)
   | (cons ?h ?t) => let rev := constr:(cons h rev) in rev_append rev t 
   end in
 rev_append (nil R) l.

 Ltac TPphi_dev add mul fv P :=
  let tl l := match l with (cons ?h ?t) => constr:(t) end in
  let rec jump j l :=
   match j with
   | xH => tl l
   | (xO ?j) => let l := jump j l in jump j l
   | (xI ?j) => let t := tl l in let l := jump j l in jump j l
   end in
  let rec mkmult r lm :=
   match lm with
   | (nil _) => constr:r
   | (cons ?h ?t) => let r := constr:(mul r h) in mkmult r t
  end in
  let mkadd_mult rP lm :=
   match lm with
   | (nil _) => constr:(add rP 1)
   | (cons ?h ?t) => let res := mkmult h t in constr:(add rP res)
   end in 
  let rec powl i x l := 
   match i with
   | xH => constr:(cons x l)
   | (xO ?i) => let l := powl i x l in powl i x l
   | (xI ?i) => let l := constr:(cons x l) in let l := powl i x l in powl i x l
   end in
  let rec add_mult_dev rP P fv lm :=
   match P with
   | (Pc ?c) => 
     match eval compute in (c ?=! cI) with
     | true => 
	let rlm := Trev lm in 
        mkadd_mult rP rlm
     | false => 
        let rlm := Trev lm in 
	let rc := eval compute in [c] in 
        let lm := constr:(cons rc rlm) in
        mkadd_mult rP lm
     end
   | (Pinj ?j ?Q) => 
      let fv := jump j fv in
      add_mult_dev rP Q fv lm
   | (PX ?P ?i ?Q) => 
     match fv with
     | (cons ?hd ?tl) =>
      let lm' := powl i hd lm in
      let rP := add_mult_dev rP P fv lm' in
      match eval compute in (Q ?== P0) with
      | true => constr:(rP)
      | false => add_mult_dev rP Q tl lm
      end
     end
   end in 
  let mkmult1 lm :=
   match lm with
   | (nil _) => eval compute in [cI]
   | (cons ?h ?t) => mkmult h t
   end in
  let rec mult_dev P fv lm :=
   match P with
   | (Pc ?c) =>
     match eval compute in (c ?=! cI) with
     | true => let rlm := Trev lm in mkmult1 rlm
     | false => 
	let rlm := Trev lm in
        let rc := eval compute in [c] in mkmult rc rlm
     end
  | (Pinj ?j ?Q) => let fv := jump j fv in mult_dev Q fv lm
  | (PX ?P ?i ?Q) =>
     match fv with
     | (cons ?hd ?tl) =>
       let lm' := powl i hd lm in
       let rP := mult_dev P fv lm' in
       match eval compute in (Q ?== P0) with
       | true => constr:(rP)
       | false => add_mult_dev rP Q tl lm
       end
     end
  end in mult_dev P fv (nil R).

 Ltac Make_ring_rewrite1 add mul fv r pe :=
  let Heq := fresh "Heq" in
  let x := fresh "x" in 
  (assert (Heq := Pphi_dev_ok fv pe);
   set (x := norm pe) in Heq;
   compute in x;
   let np := eval red in x in
   let nr := TPphi_dev add mul fv np in
   (change (req r nr) in Heq;try rewrite Heq;clear Heq;clear x)).

 Ltac Make_ring_rewrite add mul sub opp Cst_tac r :=
  let rfv := FV Cst_tac add mul sub opp r (nil R) in
  let fv := Trev rfv in
  let pe := mkPolexpr Cst_tac add mul sub opp r fv in
  Make_ring_rewrite1 add mul fv r pe.

 (* Building the generique tactic *)

 Ltac Make_ring_tac add mul sub opp req Cst_tac :=
  match goal with
  | |- (req ?r1 ?r2) =>
    let rfv := FV Cst_tac add mul sub opp (add r1 r2) (nil R) in
    let fv := Trev rfv in
    let pe1 := mkPolexpr Cst_tac add mul sub opp r1 fv in
    let pe2 := mkPolexpr Cst_tac add mul sub opp r2 fv in
    ((apply (ring_correct fv pe1 pe2);
      compute;
      exact (refl_equal true)) ||
     (Make_ring_rewrite1 add mul fv r1 pe1;
      Make_ring_rewrite1 add mul fv r2 pe2))
| |- (?r1 = ?r2) =>
    let rfv := FV Cst_tac add mul sub opp (add r1 r2) (nil R) in
    let fv := Trev rfv in
    let pe1 := mkPolexpr Cst_tac add mul sub opp r1 fv in
    let pe2 := mkPolexpr Cst_tac add mul sub opp r2 fv in
(*
    ((apply (ring_correct fv pe1 pe2);
      compute;
      exact (refl_equal true)) ||
     (Make_ring_rewrite1 add mul fv r1 pe1;
      Make_ring_rewrite1 add mul fv r2 pe2))
*)
(generalize (refl_equal pe1); generalize (refl_equal pe2);
generalize (refl_equal fv))
  | _ => fail  "not equality"
  end.
  
  (************************************************)
 (* avec des parentheses mais un peu plus efficace *)


 (**************************************************

  Fixpoint pow_mult (i:positive) (x r:R){struct i}:R :=
  match i with
  | xH => r * x
  | xO i => pow_mult i x (pow_mult i x r)
  | xI i => pow_mult i x (pow_mult i x (r * x))
  end.
 
 Definition pow_dev i x :=
  match i with
  | xH => x
  | xO i => pow_mult (Pdouble_minus_one i) x x
  | xI i => pow_mult (xO i) x x
  end.

 Lemma pow_mult_pow : forall i x r, pow_mult i x r == pow x i * r.
 Proof.
  induction i;simpl;intros.
  rewrite (IHi x (pow_mult i x (r * x)));rewrite (IHi x (r*x));rsimpl.
  mul_push x;rrefl.
  rewrite (IHi x (pow_mult i x r));rewrite (IHi x r);rsimpl.
  apply ARth.(ARmul_sym).
 Qed.

 Lemma pow_dev_pow : forall p x, pow_dev p x == pow x p.
 Proof.
  destruct p;simpl;intros.
  rewrite (pow_mult_pow p x (pow_mult p x x)).
  rewrite (pow_mult_pow p x x);rsimpl;mul_push x;rrefl.
  rewrite (pow_mult_pow (Pdouble_minus_one p) x x).
  rewrite (ARth.(ARmul_sym) (pow x (Pdouble_minus_one p)) x).
  rewrite <- (pow_Psucc x (Pdouble_minus_one p)).
  rewrite Psucc_o_double_minus_one_eq_xO;simpl; rrefl.
  rrefl.
 Qed.

 Fixpoint Pphi_dev (fv:list R) (P:Pol) {struct P} : R :=
  match P with
  | Pc c => [c]
  | Pinj j Q => Pphi_dev (jump j fv) Q
  | PX P i Q => 
    let rP := mult_dev P fv (pow_dev i (hd 0 fv)) in
    add_dev rP Q (tl fv)
  end

 with add_dev (ra:R) (P:Pol) (fv:list R) {struct P} : R :=
   match P with
  | Pc c => if c ?=! cO then ra else ra + [c] 
  | Pinj j Q => add_dev ra Q (jump j fv)
  | PX P i Q =>
    let ra := add_mult_dev ra P fv (pow_dev i (hd 0 fv)) in
    add_dev ra Q (tl fv)
  end
 
 with mult_dev (P:Pol) (fv:list R) (rm:R) {struct P} : R :=
  match P with
  | Pc c => if c ?=! cI then rm else [c]*rm
  | Pinj j Q => mult_dev Q (jump j fv) rm
  | PX P i Q =>
    let ra := mult_dev P fv (pow_mult i (hd 0 fv) rm) in
    add_mult_dev ra Q (tl fv) rm
  end

 with add_mult_dev (ra:R) (P:Pol) (fv:list R) (rm:R) {struct P} : R :=
  match P with
  | Pc c =>  if c ?=! cO then ra else ra + [c]*rm
  | Pinj j Q => add_mult_dev ra Q (jump j fv) rm
  | PX P i Q =>
    let rmP := pow_mult i (hd 0 fv) rm in
    let raP := add_mult_dev ra P fv rmP in
    add_mult_dev raP Q (tl fv) rm
  end.
 
 Lemma Pphi_add_mult_dev : forall P ra fv rm,
    add_mult_dev ra P fv rm == ra + P@fv * rm.
 Proof.
  induction P;simpl;intros.
  assert (H := CRmorph.(morph_eq) c cO).
   destruct (c ?=! cO).
    rewrite (H (refl_equal true));rewrite CRmorph.(morph0);Esimpl.
    rrefl.
  apply IHP.
  rewrite (IHP2 (add_mult_dev ra P2 fv (pow_mult p (hd 0 fv) rm)) (tl fv) rm).
  rewrite (IHP1 ra fv (pow_mult p (hd 0 fv) rm)).
  rewrite (pow_mult_pow p (hd 0 fv) rm);rsimpl.
 Qed.
  
 Lemma Pphi_add_dev : forall P ra fv, add_dev ra P fv == ra + P@fv.
 Proof.
  induction P;simpl;intros.
  assert (H := CRmorph.(morph_eq) c cO).
   destruct (c ?=! cO).
    rewrite (H (refl_equal true));rewrite CRmorph.(morph0);Esimpl.
    rrefl.
  apply IHP.
  rewrite (IHP2 (add_mult_dev ra P2 fv (pow_dev p (hd 0 fv))) (tl fv)).
  rewrite (Pphi_add_mult_dev P2 ra fv (pow_dev p (hd 0 fv))).
  rewrite  (pow_dev_pow p (hd 0 fv));rsimpl. 
 Qed.

 Lemma Pphi_mult_dev : forall P fv rm, mult_dev P fv rm == P@fv * rm.
 Proof.
  induction P;simpl;intros.
  assert (H := CRmorph.(morph_eq) c cI).
   destruct (c ?=! cI).
    rewrite (H (refl_equal true));rewrite CRmorph.(morph1);Esimpl.
    rrefl.
  apply IHP.
  rewrite (Pphi_add_mult_dev P3 
	    (mult_dev P2 fv (pow_mult p (hd 0 fv) rm)) (tl fv) rm).
  rewrite (IHP1 fv  (pow_mult p (hd 0 fv) rm)).
  rewrite (pow_mult_pow p (hd 0 fv) rm);rsimpl.
 Qed.

 Lemma Pphi_Pphi_dev : forall P fv, P@fv == Pphi_dev fv P.
 Proof.
  induction P;simpl;intros.
  rrefl. trivial.
  rewrite (Pphi_add_dev P3 (mult_dev P2 fv (pow_dev p (hd 0 fv))) (tl fv)).
  rewrite (Pphi_mult_dev P2 fv (pow_dev p (hd 0 fv))).
  rewrite (pow_dev_pow p (hd 0 fv));rsimpl.
 Qed.

 Lemma Pphi_dev_ok :  forall l pe,  PEeval l pe == Pphi_dev l (norm pe).
 Proof.
  intros l pe;rewrite <- (Pphi_Pphi_dev (norm pe) l);apply norm_ok.
 Qed.

 Ltac Trev l :=  
  let rec rev_append rev l :=
   match l with
   | (nil _) => constr:(rev)
   | (cons ?h ?t) => let rev := constr:(cons h rev) in rev_append rev t 
   end in
 rev_append (nil R) l.

 Ltac TPphi_dev add mul :=
  let tl l := match l with (cons ?h ?t) => constr:(t) end in
  let rec jump j l :=
   match j with
   | xH => tl l
   | (xO ?j) => let l := jump j l in jump j l
   | (xI ?j) => let t := tl l in let l := jump j l in jump j l
   end in
  let rec pow_mult i x r :=
   match i with
   | xH => constr:(mul r x)
   | (xO ?i) => let r := pow_mult i x r in pow_mult i x r
   | (xI ?i) => 
     let r := constr:(mul r x) in
     let r := pow_mult i x r in 
     pow_mult i x r
   end in
  let pow_dev i x :=
   match i with
   | xH => x
   | (xO ?i) => 
      let i := eval compute in (Pdouble_minus_one i) in pow_mult i x x
   | (xI ?i) => pow_mult (xO i) x x
   end in
  let rec add_mult_dev ra P fv rm :=
   match P with
   | (Pc ?c) => 
     match eval compute in (c ?=! cO) with
     | true => constr:ra 
     | _ => let rc := eval compute in [c] in constr:(add ra (mul rc rm))
     end
   | (Pinj ?j ?Q) => 
     let fv := jump j fv in add_mult_dev ra Q fv rm
   | (PX ?P ?i ?Q) =>
     match fv with 
     | (cons ?hd ?tl) =>
       let rmP := pow_mult i hd rm in
       let raP := add_mult_dev ra P fv rmP in
       add_mult_dev raP Q tl rm
     end
   end in
  let rec mult_dev P fv rm :=
   match P with
   | (Pc ?c) =>
     match eval compute in (c ?=! cI) with
     | true => constr:rm 
     | false => let rc := eval compute in [c] in constr:(mul rc rm)
     end
   | (Pinj ?j ?Q) => let fv := jump j fv in mult_dev Q fv rm
   | (PX ?P ?i ?Q) =>
     match fv with
     | (cons ?hd ?tl) =>
       let rmP := pow_mult i hd rm in
       let ra := mult_dev P fv rmP in
       add_mult_dev ra Q tl rm
     end
   end in
  let rec add_dev ra P fv :=
   match P with
   | (Pc ?c) =>
     match eval compute in (c ?=! cO) with
     | true => ra
     | false => let rc := eval compute in [c] in constr:(add ra rc)
     end
   | (Pinj ?j ?Q) => let fv := jump j fv in add_dev ra Q fv
   | (PX ?P ?i ?Q) =>
     match fv with
     | (cons ?hd ?tl) =>
       let rmP := pow_dev i hd in
       let ra := add_mult_dev ra P fv rmP in
       add_dev ra Q tl
     end
   end in
  let rec Pphi_dev fv P :=
   match P with
   | (Pc ?c) => eval compute in [c]
   | (Pinj ?j ?Q) => let fv := jump j fv in Pphi_dev fv Q
   | (PX ?P ?i ?Q) =>
     match fv with
     | (cons ?hd ?tl) =>
       let rm := pow_dev i hd in
       let rP := mult_dev P fv rm in
       add_dev rP Q tl
     end
   end in
  Pphi_dev.
 
 **************************************************************)

 (*** Un tactic pour tester la temps avec l'evaluation *)
    (** A utiliser plutot avec R *)
 Ltac Make_test add mul sub opp Cst_tac r :=
  let rfv := FV Cst_tac add mul sub opp r (nil R) in
  let fv := Trev rfv in
  let pe := mkPolexpr Cst_tac add mul sub opp r fv in
  let nr := eval compute in (Pphi_dev fv (norm pe)) in
  idtac.

End MakeRingPol. 

(** Setoid Computable Ring *)  

Module Type SetoidComputableEntries.
 Parameter R:Type.
 Parameter req : R->R->Prop. 
 Parameter Rsth : Setoid_Theory R req.   
 Parameter (rO rI : R) (radd rmul rsub: R->R->R) (ropp : R->R).
 Parameter Reqe : ring_eq_ext radd rmul ropp req.    
 Parameter Rth: ring_theory rO rI radd rmul rsub ropp req.   
 Parameter (reqb:R->R->bool).
 Parameter req_morph : forall x y, reqb x y = true -> req x y.
End SetoidComputableEntries.

Module SCE_Entries(SCE:SetoidComputableEntries).
 Import SCE.
 Definition R := R.
 Definition rO := rO.
 Definition rI := rI.
 Definition radd := radd.
 Definition rmul := rmul.
 Definition rsub := rsub.
 Definition ropp := ropp.
 Definition req := req. 
 Definition Rsth := Rsth.
 Definition Reqe := Reqe.
 Lemma ARth : almost_ring_theory rO rI radd rmul rsub ropp req.
 Proof (@Rth_ARth R rO rI radd rmul rsub ropp req Rsth Reqe Rth).

  (* Coefficients *) 
 Definition C := R.
 Definition cO := rO.
 Definition cI := rI.
 Definition cadd := radd.
 Definition cmul := rmul.
 Definition csub := rsub.
 Definition copp := ropp.
 Definition ceqb := reqb.
 Definition phi := @IDphi R.
 Lemma CRmorph : ring_morph rO rI radd rmul rsub ropp req
                            cO cI cadd cmul csub copp ceqb phi.
 Proof (@IDmorph R rO rI radd rmul rsub ropp req Rsth reqb req_morph).
End SCE_Entries. 

(** Computable Ring *)

Module Type ComputableEntries.
 Parameter R:Type.
 Parameter (rO rI : R) (radd rmul rsub: R->R->R) (ropp : R->R).
 Parameter Rth: ring_theory rO rI radd rmul rsub ropp (@eq R).   
 Parameter (reqb:R->R->bool).
 Parameter req_morph : forall x y, reqb x y = true -> x = y.
End ComputableEntries.

Module CE_Entries(CE:ComputableEntries).
 Import CE.
 Definition R := R.
 Definition rO := rO.
 Definition rI := rI.
 Definition radd := radd.
 Definition rmul := rmul.
 Definition rsub := rsub.
 Definition ropp := ropp.
 Definition req := (@eq R). 
 Lemma Rsth : Setoid_Theory R req.  
 Proof (Eqsth R).
 Lemma Reqe : ring_eq_ext radd rmul ropp req.
 Proof (@Eq_ext R radd rmul ropp).
 Lemma ARth : almost_ring_theory rO rI radd rmul rsub ropp req.
 Proof (@Rth_ARth R rO rI radd rmul rsub ropp req Rsth Reqe Rth).

  (* Coefficients *) 
 Definition C := R.
 Definition cO := rO.
 Definition cI := rI.
 Definition cadd := radd.
 Definition cmul := rmul.
 Definition csub := rsub.
 Definition copp := ropp.
 Definition ceqb := reqb.
 Definition phi := @IDphi R.
 Lemma CRmorph : ring_morph rO rI radd rmul rsub ropp req
                            cO cI cadd cmul csub copp ceqb phi.
 Proof (@IDmorph R rO rI radd rmul rsub ropp req Rsth reqb req_morph).
End CE_Entries. 
	   
(** Setoid Computable Semi Ring *)	  
 
Module Type SetoidComputableSemiEntries.
 Parameter R:Type.
 Parameter req : R->R->Prop. 
 Parameter Rsth : Setoid_Theory R req.   
 Parameter (rO rI : R) (radd rmul: R->R->R).
 Parameter SReqe : sring_eq_ext radd rmul req.    
 Parameter SRth: semi_ring_theory rO rI radd rmul req.   
 Parameter (reqb:R->R->bool).
 Parameter req_morph : forall x y, reqb x y = true -> req x y.
End SetoidComputableSemiEntries.

Module SCSE_Entries(SCSE:SetoidComputableSemiEntries).
 Import SCSE.
 Definition R := R.
 Definition rO := rO.
 Definition rI := rI.
 Definition radd := radd.
 Definition rmul := rmul.
 Definition rsub := (@SRsub R radd).
 Definition ropp := (@SRopp R).
 Definition req := req. 
 Definition Rsth := Rsth.
 Definition Reqe := (@SReqe_Reqe R radd rmul req SReqe).
 Lemma ARth : almost_ring_theory rO rI radd rmul rsub ropp req.
 Proof (@SRth_ARth R rO rI radd rmul req Rsth SRth).

  (* Coefficients *) 
 Definition C := R.
 Definition cO := rO.
 Definition cI := rI.
 Definition cadd := radd.
 Definition cmul := rmul.
 Definition csub := rsub.
 Definition copp := ropp.
 Definition ceqb := reqb.
 Definition phi := @IDphi R.
 Lemma CRmorph : ring_morph rO rI radd rmul rsub ropp req
                            cO cI cadd cmul csub copp ceqb phi.
 Proof (@SRIDmorph R rO rI radd rmul req Rsth reqb req_morph).
End SCSE_Entries. 

(* Computable Semi Ring *)		 
Module Type ComputableSemiEntries.
 Parameter R:Type.
 Parameter (rO rI : R) (radd rmul: R->R->R).
 Parameter SRth: semi_ring_theory rO rI radd rmul (@eq R).   
 Parameter (reqb:R->R->bool).
 Parameter req_morph : forall x y, reqb x y = true -> x = y.
End ComputableSemiEntries.

Module CSE_Entries(CSE:ComputableSemiEntries).
 Import CSE.
 Definition R := R.
 Definition rO := rO.
 Definition rI := rI.
 Definition radd := radd.
 Definition rmul := rmul.
 Definition rsub := (@SRsub R radd).
 Definition ropp := (@SRopp R).
 Definition req := @eq R.
 Lemma Rsth : Setoid_Theory R req.
 Proof (Eqsth R).
 Definition Reqe := (@SReqe_Reqe R radd rmul req (@Eq_s_ext R radd rmul)).
 Lemma ARth : almost_ring_theory rO rI radd rmul rsub ropp req.
 Proof (@SRth_ARth R rO rI radd rmul req Rsth SRth).

  (* Coefficients *) 
 Definition C := R.
 Definition cO := rO.
 Definition cI := rI.
 Definition cadd := radd.
 Definition cmul := rmul.
 Definition csub := rsub.
 Definition copp := ropp.
 Definition ceqb := reqb.
 Definition phi := @IDphi R.
 Lemma CRmorph : ring_morph rO rI radd rmul rsub ropp req
                            cO cI cadd cmul csub copp ceqb phi.
 Proof (@SRIDmorph R rO rI radd rmul req Rsth reqb req_morph).
End CSE_Entries. 

(** Setoid Abstract Ring *)  
Require Import ZArith.
Open Scope Z_scope.
		 
Module Type SetoidAbstractEntries.
 Parameter R:Type.
 Parameter req : R->R->Prop. 
 Parameter Rsth : Setoid_Theory R req.   
 Parameter (rO rI : R) (radd rmul rsub: R->R->R) (ropp : R->R).
 Parameter Reqe : ring_eq_ext radd rmul ropp req.    
 Parameter Rth: ring_theory rO rI radd rmul rsub ropp req.   
End SetoidAbstractEntries.

Module SAE_Entries(SAE:SetoidAbstractEntries).
 Import SAE.
 Definition R := R.
 Definition rO := rO.
 Definition rI := rI.
 Definition radd := radd.
 Definition rmul := rmul.
 Definition rsub := rsub.
 Definition ropp := ropp.
 Definition req := req. 
 Definition Rsth := Rsth.
 Definition Reqe := Reqe.
 Lemma ARth : almost_ring_theory rO rI radd rmul rsub ropp req.
 Proof (@Rth_ARth R rO rI radd rmul rsub ropp req Rsth Reqe Rth).

  (* Coefficients *) 
 Definition C := Z.
 Definition cO := 0.
 Definition cI := 1.
 Definition cadd := Zplus.
 Definition cmul := Zmult.
 Definition csub := Zminus.
 Definition copp := Zopp.
 Definition ceqb := Zeq_bool.
 Definition phi := @gen_phiZ R rO rI radd rmul ropp.
 Lemma CRmorph : ring_morph rO rI radd rmul rsub ropp req
                            cO cI cadd cmul csub copp ceqb phi.
 Proof (@gen_phiZ_morph R rO rI radd rmul rsub ropp req Rsth Reqe Rth).
End SAE_Entries. 

(** Computable Ring *)

Module Type AbstractEntries.
 Parameter R:Type.
 Parameter (rO rI : R) (radd rmul rsub: R->R->R) (ropp : R->R).
 Parameter Rth: ring_theory rO rI radd rmul rsub ropp (@eq R).   
End AbstractEntries.

Module AE_Entries(CE:AbstractEntries).
 Import CE.
 Definition R := R.
 Definition rO := rO.
 Definition rI := rI.
 Definition radd := radd.
 Definition rmul := rmul.
 Definition rsub := rsub.
 Definition ropp := ropp.
 Definition req := (@eq R). 
 Lemma Rsth : Setoid_Theory R req.  
 Proof (Eqsth R).
 Lemma Reqe : ring_eq_ext radd rmul ropp req.
 Proof (@Eq_ext R radd rmul ropp).
 Lemma ARth : almost_ring_theory rO rI radd rmul rsub ropp req.
 Proof (@Rth_ARth R rO rI radd rmul rsub ropp req Rsth Reqe Rth).

  (* Coefficients *) 
 Definition C := Z.
 Definition cO := 0.
 Definition cI := 1.
 Definition cadd := Zplus.
 Definition cmul := Zmult.
 Definition csub := Zminus.
 Definition copp := Zopp.
 Definition ceqb := Zeq_bool.
 Definition phi := @gen_phiZ R rO rI radd rmul ropp.
 Lemma CRmorph : ring_morph rO rI radd rmul rsub ropp req
                            cO cI cadd cmul csub copp ceqb phi.
 Proof (@gen_phiZ_morph R rO rI radd rmul rsub ropp req Rsth Reqe Rth).
End AE_Entries. 
	   
(** Setoid Abstract Semi Ring *)	  

Require Import NArith.
Open Scope N_scope.
 
Module Type SetoidAbstractSemiEntries.
 Parameter R:Type.
 Parameter req : R->R->Prop. 
 Parameter Rsth : Setoid_Theory R req.   
 Parameter (rO rI : R) (radd rmul: R->R->R).
 Parameter SReqe : sring_eq_ext radd rmul req.    
 Parameter SRth: semi_ring_theory rO rI radd rmul req.   
End SetoidAbstractSemiEntries.

Module SASE_Entries(SASE:SetoidAbstractSemiEntries).
 Import SASE.
 Definition R := R.
 Definition rO := rO.
 Definition rI := rI.
 Definition radd := radd.
 Definition rmul := rmul.
 Definition rsub := (@SRsub R radd).
 Definition ropp := (@SRopp R).
 Definition req := req. 
 Definition Rsth := Rsth.
 Definition Reqe := (@SReqe_Reqe R radd rmul req SReqe).
 Lemma ARth : almost_ring_theory rO rI radd rmul rsub ropp req.
 Proof (@SRth_ARth R rO rI radd rmul req Rsth SRth).

  (* Coefficients *) 
 Definition C := N.
 Definition cO := 0.
 Definition cI := 1.
 Definition cadd := Nplus.
 Definition cmul := Nmult.
 Definition csub := (@SRsub C cadd).
 Definition copp := (@SRopp C).
 Definition ceqb := Neq_bool.
 Definition phi := @gen_phiN R rO rI radd rmul.
 Lemma CRmorph : ring_morph rO rI radd rmul rsub ropp req
                            cO cI cadd cmul csub copp ceqb phi.
 Proof (@gen_phiN_morph R rO rI radd rmul req Rsth SReqe SRth).

End SASE_Entries. 

(* Abstract Semi Ring *)		 
Module Type AbstractSemiEntries.
 Parameter R:Type.
 Parameter (rO rI : R) (radd rmul: R->R->R).
 Parameter SRth: semi_ring_theory rO rI radd rmul (@eq R).   
End AbstractSemiEntries.

Module ASE_Entries(ASE:AbstractSemiEntries).
 Import ASE.
 Definition R := R.
 Definition rO := rO.
 Definition rI := rI.
 Definition radd := radd.
 Definition rmul := rmul.
 Definition rsub := (@SRsub R radd).
 Definition ropp := (@SRopp R).
 Definition req := @eq R.
 Lemma Rsth : Setoid_Theory R req.
 Proof (Eqsth R).
 Definition Reqe := (@SReqe_Reqe R radd rmul req (@Eq_s_ext R radd rmul)).
 Lemma ARth : almost_ring_theory rO rI radd rmul rsub ropp req.
 Proof (@SRth_ARth R rO rI radd rmul req Rsth SRth).

  (* Coefficients *) 
 Definition C := N.
 Definition cO := 0.
 Definition cI := 1.
 Definition cadd := Nplus.
 Definition cmul := Nmult.
 Definition csub := (@SRsub C cadd).
 Definition copp := (@SRopp C).
 Definition ceqb := Neq_bool.
 Definition phi := @gen_phiN R rO rI radd rmul.
 Lemma CRmorph : ring_morph rO rI radd rmul rsub ropp req
                            cO cI cadd cmul csub copp ceqb phi.
 Proof 
  (@gen_phiN_morph R rO rI radd rmul req Rsth (@Eq_s_ext R radd rmul) SRth).

End ASE_Entries. 

  
