Require Export Reals.
Require Export Field.
Require Export ZArith.
Require Export Omega.
Require Export ZArithRing.
Require Export Wellfounded.
Require Export Zwf.
Require Export Fourier.
Require Export List.

Open Scope Z_scope.

CoInductive stream (A:Set) : Set :=
  Cons : A -> stream A -> stream A.

Implicit Arguments Cons.

Infix "::" := Cons :stream_scope.

Open Scope stream_scope.

Inductive idigit : Set :=  L | R | C.

CoInductive represents : stream idigit ->
                  Rdefinitions.R -> Prop :=
  reprL : forall s r, represents s r -> (-1 <= r <= 1)%R ->
           represents (Cons L s) (r/2-1/2)
| reprR : forall s r, represents s r -> (-1 <= r <= 1)%R ->
           represents (Cons R s) (r/2+1/2)
| reprC : forall s r, represents s r -> (-1 <= r <= 1)%R ->
           represents (Cons C s) (r/2).

(* returns the interval represented by the n first digits of the stream,
   the value is given by (a, b, k) so that the interval actually is
   (a/(2**k), b/(2**k)) *)
Fixpoint bounds (n:nat) (s:stream idigit) {struct n} : Z*Z*Z :=
  match n with
    0%nat => (-1,1,0)
  | S p =>
    let (d,tl) := s in
    let (p,e) := bounds p tl in
    let (a,b) := p in
    match d with
      L => (a - Zpower 2 e,b - Zpower 2 e,e+1)
    | R => (a + Zpower 2 e,b + Zpower 2 e,e+1)
    | C => (a, b, e+1)
    end
  end.

(* constructs the infinite stream representing a rational number.
   This function works even if the rational number is not between
   0 and 1, but the value is meaningless. *)
CoFixpoint rat_to_stream (a b:Z) : stream idigit :=
  if Z_le_gt_dec 0 (a*b) then
    Cons R (rat_to_stream (2*a-b) b)
  else
    Cons L (rat_to_stream (2*a+b) b).

(* A few constants. *)

CoFixpoint one : stream idigit := R::one.

CoFixpoint sub_one : stream idigit := L::sub_one.

CoFixpoint zero : stream idigit := C::zero.

CoFixpoint mult2 (x:stream idigit) : stream idigit :=
  match x with
    | L::x' =>
      match x' with 
	| L::_ => sub_one
	| C::x'' => L::mult2 (L::x'')
	| R::x'' => L::x''
      end 
    | C::x' => x'
    | R::x' => 
      match x' with 
	| L::x'' => R::x''
	| C::x'' => R::mult2 (R::x'')
	| R::_   => one
      end
  end.

(* half_sum two numbers. *)

Definition help_type (d1 d2:stream idigit) := stream idigit.
Definition help_type2 (d1 d2:stream idigit) (d3:idigit) := stream idigit.

Ltac stream_length f :=
  match f with
    Cons ?x ?f' => let v := stream_length f' in constr:(S v)
  | _ => constr:0%nat
  end.

Ltac stream_tail f :=
 match f with Cons ?x ?f' => f' | _ => f end.

Ltac stream_end f :=
 match f with Cons ?x ?f' => stream_end f' | _ => f end.

Ltac mk_cases :=
  match goal with |- help_type ?a ?b =>
    let n1 := (stream_length a) in
    let n2 := (stream_length b) in
    let tl1 := (stream_end a) in
    let tl2 := (stream_end b) in
    let n3 := eval compute in (n1 - n2)%nat in
    let n4 := eval compute in (4 - (n1 + n2))%nat in
    match n4 with
      0 => idtac "mk_cases failing"; fail
    | (S _) =>
      match n3 with
        0%nat => case tl1; intros [ | | ]; intro
      | (S _) => case tl2; intros [ | | ]; intro
      end
    end
  end.

Ltac den s :=
  let n := stream_length s in
  let b := eval compute in (bounds n s) in
  let r := eval compute in (2^(snd b))%Z in
  r.

Ltac ub s :=
  let n := stream_length s in
  let b := eval compute in (bounds n s) in
  let r := eval compute in (snd (fst b)) in
  r.

Ltac lb s :=
  let n := stream_length s in
  let b := eval compute in (bounds n s) in
  let r := eval compute in (fst (fst b)) in
  r.

Ltac produce_add_cases :=
 match goal with |- help_type ?u ?v =>
  let bb1 := lb u in let bb2 := lb v in
  let tb1 := ub u in let tb2 := ub v in
  let d1 := den u in let d2 := den v in
  let test1 := eval compute in
    (bb1*d2 + bb2*d1 ?= 0)%Z in
  let test2 := eval compute in
    (tb1*d2 + tb2*d1 ?= 0)%Z in
  let test3 := eval compute in
    (tb1*d2 + tb2*d1 ?= d1*d2)%Z in
  let test4 := eval compute in 
    (bb1*d2 + bb2*d1 ?= -d1*d2)%Z in
  match test1 with 
    Gt => apply (Cons R); change (help_type2 u v R)
  | Eq => apply (Cons R); change (help_type2 u v R)
  | Lt =>
    match test2 with
      Lt => apply (Cons L); change (help_type2 u v L)
    | Eq => apply (Cons L); change (help_type2 u v L)
    | Gt =>
      match test3 with
        Gt => try (mk_cases; produce_add_cases)
      | _ =>
        match test4 with
          Lt => try (mk_cases; produce_add_cases)
        | _ => apply (Cons C); try change (help_type2 u v C)
        end
      end
    end
  end
end.

Definition lbf (d:idigit) : Z :=
  match d with L => -1 | R => 1 | C => 0 end.

Definition compute_borrow (i1 i2 r :idigit) : Z :=
  lbf i1 + lbf i2 - 2*lbf r.

Ltac mk_borrow half_sum v s1 s2 :=
  let mv := constr:(v,s1,s2) in
  match mv with
  | (-1, R::?tl1, ?tl2) => exact (half_sum (L::tl1) tl2)
  | (-1, ?tl1, R::?tl2) => exact (half_sum tl1 (L::tl2))
  | (-1, C::?tl1, C::?tl2) => exact (half_sum (L::tl1) (L::tl2))
  | (1, L::?tl1, ?tl2) => exact (half_sum (R::tl1) tl2)
  | (1, ?tl1, L::?tl2) => exact (half_sum tl1 (R::tl2))
  | (1, C::?tl1, C::?tl2) => exact (half_sum (R::tl1) (R::tl2))
  end.

Definition half_sum (x y:stream idigit) : stream idigit.
cofix.
intros x y; change (help_type x y).
produce_add_cases;
match goal with |- help_type2 (?x::?tl1) (?y::?tl2) ?z =>
 let r := eval compute in (compute_borrow x y z) in
 match r with 0 => exact (half_sum tl1 tl2)
   | _ => mk_borrow half_sum r tl1 tl2
 end
end.
Defined.

Definition add (x y:stream idigit) : stream idigit :=
  mult2(half_sum x y).

Infix "+" := add : stream_scope.
Delimit Scope stream_scope with S.

CoFixpoint Sopp (s:stream idigit) : stream idigit :=
  match s with 
    (d::s') => Cons (match d with L => R| C => C | R => L end) (Sopp s')
  end.

CoFixpoint half_sub (x y:stream idigit) : stream idigit :=
  match x with 
    | L::x' => 
      match y with
	| L::y' => C::(half_sub x' y')
	| C::y' => 
	  match y' with
	    | L::y'' => C::(half_sub x' (R::y''))
	    | C::y'' => 
	      match x' with
		| L::x'' => L::(half_sub (C::x'') (L::y''))
		| C::x'' => C::(half_sub (L::x'') (R::y''))
		| R::x'' => C::(half_sub (L::x'') y')
	      end
	    | R::y'' => L::(half_sub x' (L::y''))
	  end
	| R::y' => L::(half_sub x' y')
      end
    | C::x' => 
      match y with
	| L::y' => 
	  match x' with
	    | L::x'' => C::(half_sub (R::x'') y')
	    | C::x'' => 
	      match y' with
		| L::y'' => R::(half_sub (L::x'') (C::y''))
		| C::y'' => C::(half_sub (R::x'') (L::y''))
		| R::y'' => C::(half_sub x' (L::y''))
	      end
	    | R::x'' => R::(half_sub (L::x'') y')
	  end
	| C::y' => C::(half_sub x' y')
	| R::y' => 
	  match x' with
	    | L::x'' => L::(half_sub (R::x'') y')
	    | C::x'' => 
	      match y' with
		| L::y'' => C::(half_sub x' (R::y''))
		| C::y'' => C::(half_sub (L::x'') (R::y''))
		| R::y'' => L::(half_sub (R::x'') (C::y''))
	      end
	    | R::x'' => C::(half_sub (L::x'') y')
	  end
      end
    | R::x' => 
      match y with
	| L::y' => R::(half_sub x' y')
	| C::y' => 
	  match y' with
	    | L::y'' => R::(half_sub x' (R::y''))
	    | C::y'' => 
	      match x' with
		| L::x'' => C::(half_sub (R::x'') y')
		| C::x'' => C::(half_sub (R::x'') (L::y''))
		| R::x'' => R::(half_sub (C::x'') (R::y''))
	      end
	    | R::y'' => C::(half_sub x' (L::y''))
	  end
	| R::y' => C::(half_sub x' y')
      end
  end.

Definition sub (x y : stream idigit) : stream idigit := 
  mult2 (half_sub x y). 

Infix "-" := sub : stream_scope.

Definition series_body_pos (A:Set)
  (series : stream idigit -> A -> stream idigit)
  (x : stream idigit) (a:A) : stream idigit :=
  let (dig,x'') := x in
    match dig with
      R =>  R::series x'' a
      | L => match x'' with
               R::x3 => C::series (L::x3) a
               | _ => L::series x'' a
             end
      | C => match x'' with
               R::x3 => R::series (L::x3) a
               | _ => C::series x'' a
             end
    end.

Definition series_body (A:Set)
  (series : stream idigit -> A -> stream idigit) 
  (x : stream idigit) (a:A) : stream idigit :=
  let (dig,x') := x in
    match dig with 
      | L => 
	match x' with
          | R::x'' => 
	    match x'' with 
	      | R::x3 => C::series (L::R::x3) a
              | _ => L::series x' a
            end
          | _ => L::series x' a
	end
      | C => 
	match x' with
          | L::x'' => 
	    match x'' with 
	      | L::x3 => L::series (R::L::x3) a
              | _ => C::series x' a
            end
          | R::x'' => 
	    match x'' with 
	      | R::x3 => R::series (L::R::x3) a
              | _ => C::series x' a
            end
          | _ => C::series x' a
        end
      | R =>  
	match x' with
          | L::x'' => 
	    match x'' with 
	      | L::x3 => C::series (R::L::x3) a
              | _ => R::series x' a
            end
          | _ => R::series x' a
	end
    end.

(* compute x + uw *)
CoFixpoint mult_a (x:stream idigit) (p:stream idigit*stream idigit) : stream idigit :=
  let (u,v) := p in
  match u with
    | L::u' => 
      match u' with
	| L::_ => series_body _ mult_a (x-(C::v)) (u',v)
	| C::u'' => series_body _ mult_a (x-(C::C::v)) (L::u'',v)
	| R::u'' => series_body _ mult_a x (L::u'',v)
      end
    | C::u' => series_body _ mult_a x (u',v)
    | R::u' => 
      match u' with
	| L::u'' => series_body _ mult_a x (R::u'',v)
	| C::u'' => series_body _ mult_a (x+(C::C::v)) (R::u'',v)
	| R::u'' => series_body _ mult_a (x+(C::v)) (u',v)
      end
  end.

(*
CoFixpoint mult_a (x:stream idigit) (p:stream idigit*stream idigit) 
  : stream idigit :=
  let (u,v) := p in
  match u with
    L::u' => series_body _ mult_a (x-(C::v)) (u',v)
  | C::u' => series_body _ mult_a x (u',v)
  | R::u' => series_body _ mult_a (x+(C::v)) (u',v)
  end.
*)

Definition mult (x y:stream idigit) : stream idigit :=
  mult2(mult2(mult_a zero (x,C::C::y))).

Infix "*" := mult : stream_scope.

Inductive triple : Set :=
  mk_triple : stream idigit -> Z -> Z -> triple.

CoFixpoint e_series (v:stream idigit)(s :Z*Z*Z) :stream idigit :=
  let (aux, fact_nm1) := s in let (y, n) := aux in
  let (v', n', fact_nm1') := 
     if Z_le_gt_dec (8*y) (fact_nm1*(n-1))%Z then
        mk_triple v n fact_nm1
     else
        mk_triple (v + (rat_to_stream y (fact_nm1 * n)))
          (n+1) (fact_nm1*n) in
     series_body_pos  _ e_series v' (2*y, n', fact_nm1')%Z.

Definition number_e_minus2 : stream idigit :=
  e_series (rat_to_stream 1 2) (1, 3, 2).

Definition number_e_minus2_square : stream idigit :=
  number_e_minus2 * number_e_minus2.

CoFixpoint arctan_1_n_aux (v:stream idigit)(s :Z*Z*Z*Z) : stream idigit :=
  let (aux1, a) := s in
    let (aux2, n) := aux1 in
      let (y, i) := aux2 in
	if (Z_le_gt_dec a (4*i-3)) then
	  series_body_pos _ arctan_1_n_aux v (2*y, i, n, a+1)%Z
	else
	  let c := (n^2*(4*i+3)-4*i-1)%Z in 
	    let d := (n^(4*i+3)*(4*i+1)*(4*i+3))%Z in
	      let v' := rat_to_stream (y * c)%Z d in
		series_body_pos _ arctan_1_n_aux (v+v')%S (2*y, (i+1), n, a+1)%Z.

Definition arctan_1_n (n:Z): stream idigit :=
  arctan_1_n_aux (rat_to_stream  (3*n^2-1) (3*n^3)) (1, 1, n, 0).

Definition pi_div_4 :=
  arctan_1_n 2 + arctan_1_n 3.
(*
Time Eval compute in (bounds 30 number_e_minus2).
*)

Time Eval compute in (bounds  30 pi_div_4).

