Require Export Reals.
Require Export Field.
Require Export ZArith.
Require Export Omega.
Require Export ZArithRing.
Require Export Wellfounded.
Require Export Zwf.
Require Export Fourier.
Require Export List.

Open Scope Z_scope.

CoInductive stream (A:Set) : Set :=
  Cons : A -> stream A -> stream A.

Implicit Arguments Cons.

Infix "::" := Cons :stream_scope.

Open Scope stream_scope.

Inductive idigit : Set :=  L | R | C.

CoInductive represents : stream idigit ->
                  Rdefinitions.R -> Prop :=
  reprL : forall s r, represents s r -> (0 <= r <= 1)%R ->
           represents (Cons L s) (r/2)
| reprR : forall s r, represents s r -> (0 <= r <= 1)%R ->
           represents (Cons R s) ((r+1)/2)
| reprC : forall s r, represents s r -> (0 <= r <= 1)%R ->
           represents (Cons C s) ((2*r+1)/4).

(* returns the interval represented by the n first digits of the stream,
   the value is given by (a, b, k) so that the interval actually is
   (a/(2**k), b/(2**k)) *)

Fixpoint bounds (n:nat) (s:stream idigit) {struct n} : Z*Z*Z :=
  match n with
    0%nat => (0,1,0)
  | S p =>
    let (d,tl) := s in
    let (p,e) := bounds p tl in
    let (a,b) := p in
    match d with
      L => (a,b,e+1)
    | R => (a+Zpower 2 e,b+Zpower 2 e,e+1)
    | C => (2*a+Zpower 2 e, 2*b+Zpower 2 e, e+2)
    end
  end.

(* constructs the infinite stream representing a rational number.
   This function works even if the rational number is not between
   0 and 1, but the value is meaningless. *)

CoFixpoint rat_to_stream (a b:Z) : stream idigit :=
  if Z_le_gt_dec (2*a) b then
    Cons L (rat_to_stream (2*a) b)
  else
    Cons R (rat_to_stream (2*a-b) b).

(* A few constants. *)

CoFixpoint one : stream idigit := R::one.

CoFixpoint zero : stream idigit := L::zero.

Definition half : stream idigit := R::zero.

CoFixpoint third : stream idigit := L::R::third.

Definition sixth : stream idigit := L::third.

(* half_sum two numbers. *)

Definition help_type (d1 d2:stream idigit) := stream idigit.
Definition help_type2 (d1 d2:stream idigit)(d3:idigit) := stream idigit.

Ltac stream_length f :=
  match f with
    Cons ?x ?f' => let v := stream_length f' in constr:(S v)
  | _ => constr:0%nat
  end.

Ltac stream_tail f :=
 match f with Cons ?x ?f' => f' | _ => f end.

Ltac stream_end f :=
 match f with Cons ?x ?f' => stream_end f' | _ => f end.


Ltac mk_cases :=
  match goal with |- help_type ?a ?b =>
    let n1 := (stream_length a) in
    let n2 := (stream_length b) in
    let tl1 := (stream_end a) in
    let tl2 := (stream_end b) in
    let n3 := eval compute in (n1 - n2)%nat in
    let n4 := eval compute in (4 - (n1 + n2))%nat in
    match n4 with
      0 => idtac "mk_cases failing"; fail
    | (S _) =>
      match n3 with
        0%nat => case tl1; intros [ | | ]; intro
      | (S _) => case tl2; intros [ | | ]; intro
      end
    end
  end.

Ltac den s :=
  let n := stream_length s in
  let b := eval compute in (bounds n s) in
  let r := eval compute in (2^(snd b))%Z in
  r.

Ltac ub s :=
  let n := stream_length s in
  let b := eval compute in (bounds n s) in
  let r := eval compute in (snd (fst b)) in
  r.

Ltac lb s :=
  let n := stream_length s in
  let b := eval compute in (bounds n s) in
  let r := eval compute in (fst (fst b)) in
  r.

Ltac produce_add_cases :=
 match goal with |- help_type ?u ?v =>
  let bb1 := lb u in let bb2 := lb v in
  let tb1 := ub u in let tb2 := ub v in
  let d1 := den u in let d2 := den v in
  let test1 := eval compute in
    (bb1*d2 + bb2*d1 ?= d1*d2)%Z in
  let test2 := eval compute in
    (tb1*d2 + tb2*d1 ?= d1*d2)%Z in
  let test3 := eval compute in
    (2*(tb1*d2 + tb2*d1) ?= (3*(d1*d2)))%Z in
  let test4 := eval compute in 
    (2*(bb1*d2 + bb2*d1) ?= d1*d2)%Z in
  match test1 with 
    Gt => apply (Cons R); change (help_type2 u v R)
  | Eq => apply (Cons R); change (help_type2 u v R)
  | Lt =>
    match test2 with
      Lt => apply (Cons L); change (help_type2 u v L)
    | Eq => apply (Cons L); change (help_type2 u v L)
    | Gt =>
      match test3 with
        Gt => try (mk_cases; produce_add_cases)
      | _ =>
        match test4 with
          Lt => try (mk_cases; produce_add_cases)
        | _ => apply (Cons C); try change (help_type2 u v C)
        end
      end
    end
  end
end.

CoFixpoint mult2 (x:stream idigit) : stream idigit :=
  match x with
    R::_ => one
  | L::x' => x'
  | C::x' => R::mult2 x'
  end.

Definition lbf (d:idigit) : Z :=
  match d with L => 0 | R => 2 | C => 1 end.


Definition compute_borrow (i1 i2 r :idigit) : Z :=
  lbf i1 + lbf i2 - 2*lbf r.

Ltac mk_borrow half_sum v s1 s2 :=
  let mv := constr:(v,s1,s2) in
  match mv with
  | (-1, R::?tl1, ?tl2) => exact (half_sum (L::tl1) tl2)
  | (-1, ?tl1, R::?tl2) => exact (half_sum tl1 (L::tl2))
  | (-1, C::?tl1, C::?tl2) => exact (half_sum (L::tl1) (L::tl2))
  | (1, L::?tl1, ?tl2) => exact (half_sum (R::tl1) tl2)
  | (1, ?tl1, L::?tl2) => exact (half_sum tl1 (R::tl2))
  | (1, C::?tl1, C::?tl2) => exact (half_sum (R::tl1) (R::tl2))
  end.

Definition half_sum (x y:stream idigit) : stream idigit.
cofix.
intros x y; change (help_type x y).
produce_add_cases;
match goal with |- help_type2 (?x::?tl1) (?y::?tl2) ?z =>
 let r := eval compute in (compute_borrow x y z) in
 match r with 0 => exact (half_sum tl1 tl2)
   | _ => mk_borrow half_sum r tl1 tl2
 end
end.
Defined.


Definition add (x y:stream idigit) : stream idigit :=
  mult2(half_sum x y).

Infix "+" := add : stream_scope.
Delimit Scope stream_scope with S.

CoFixpoint sub_aux (x y:stream idigit) : stream idigit :=
  match x with 
    | L::x' => 
      match y with
	| L::y' => C::(sub_aux x' y')
	| C::y' => 
	  match y' with
	    | L::y'' => C::(sub_aux x' (R::y''))
	    | C::y'' => 
	      match x' with
		| L::x'' => L::(sub_aux (C::x'') (L::y''))
		| C::x'' => C::(sub_aux (L::x'') (R::y''))
		| R::x'' => C::(sub_aux (L::x'') y')
	      end
	    | R::y'' => L::(sub_aux x' (L::y''))
	  end
	| R::y' => L::(sub_aux x' y')
      end
    | C::x' => 
      match y with
	| L::y' => 
	  match x' with
	    | L::x'' => C::(sub_aux (R::x'') y')
	    | C::x'' => 
	      match y' with
		| L::y'' => R::(sub_aux (L::x'') (C::y''))
		| C::y'' => C::(sub_aux (R::x'') (L::y''))
		| R::y'' => C::(sub_aux x' (L::y''))
	      end
	    | R::x'' => R::(sub_aux (L::x'') y')
	  end
	| C::y' => C::(sub_aux x' y')
	| R::y' => 
	  match x' with
	    | L::x'' => L::(sub_aux (R::x'') y')
	    | C::x'' => 
	      match y' with
		| L::y'' => C::(sub_aux x' (R::y''))
		| C::y'' => C::(sub_aux (L::x'') (R::y''))
		| R::y'' => L::(sub_aux (R::x'') (C::y''))
	      end
	    | R::x'' => C::(sub_aux (L::x'') y')
	  end
      end
    | R::x' => 
      match y with
	| L::y' => R::(sub_aux x' y')
	| C::y' => 
	  match y' with
	    | L::y'' => R::(sub_aux x' (R::y''))
	    | C::y'' => 
	      match x' with
		| L::x'' => C::(sub_aux (R::x'') y')
		| C::x'' => C::(sub_aux (R::x'') (L::y''))
		| R::x'' => R::(sub_aux (C::x'') (R::y''))
	      end
	    | R::y'' => C::(sub_aux x' (L::y''))
	  end
	| R::y' => C::(sub_aux x' y')
      end
  end.

CoFixpoint mult2_sub_one (x: stream idigit):stream idigit :=
  match x with 
    | L::_ => zero
    | C::s => L::mult2_sub_one s
    | R::s => s
  end.

Definition sub (x y : stream idigit) : stream idigit := 
  mult2_sub_one (sub_aux x y).

Infix "-" := sub : stream_scope.

Definition series_body (A:Set)
  (series : stream idigit -> A -> stream idigit) 
  (x : stream idigit) (a:A) : stream idigit :=
 let (dig,x'') := x in
  match dig with
    R =>  R::series x'' a
  | L => match x'' with
           R::x3 => C::series (L::x3) a
         | _ => L::series x'' a
         end
  | C => match x'' with
           R::x3 => R::series (L::x3) a
         | _ => C::series x'' a
         end
  end.

(* compute x + uw/4 *)
CoFixpoint mult_a (x:stream idigit) (p:stream idigit*stream idigit) 
  : stream idigit :=
  let (u,v) := p in
  match u with
    L::u' => series_body _ mult_a x (u',v)
  | C::u' => series_body _ mult_a (x+(L::L::v)) (u',v)
  | R::u' => series_body _ mult_a (x+(L::v)) (u',v)
  end.

Definition mult (x y:stream idigit) : stream idigit :=
  mult2(mult2(mult_a zero (x,L::L::y))).

Infix "*" := mult : stream_scope.

Inductive triple : Set :=
  mk_triple : stream idigit -> Z -> Z -> triple.

CoFixpoint e_series (v:stream idigit)(s :Z*Z*Z) :stream idigit :=
  let (aux, fact_nm1) := s in let (y, n) := aux in
  let (v', n', fact_nm1') := 
     if Z_le_gt_dec (8*y) (fact_nm1*(n-1))%Z then
        mk_triple v n fact_nm1
     else
        mk_triple (v + (rat_to_stream y (fact_nm1 * n)))
          (n+1) (fact_nm1*n) in
     series_body  _ e_series v' (2*y, n', fact_nm1')%Z.

Definition number_e_minus2 : stream idigit :=
  e_series (rat_to_stream 1 2) (1, 3, 2).

Definition number_e_minus2_square : stream idigit :=
  number_e_minus2 * number_e_minus2.

CoFixpoint arctan_1_n_aux (v:stream idigit)(s :Z*Z*Z*Z) : stream idigit :=
  let (aux1, a) := s in
    let (aux2, n) := aux1 in
      let (y, i) := aux2 in
	if (Z_le_gt_dec a (4*i-3)) then
	  series_body _ arctan_1_n_aux v (2*y, i, n, a+1)%Z
	else
	  let c := (n^2*(4*i+3)-4*i-1)%Z in 
	    let d := (n^(4*i+3)*(4*i+1)*(4*i+3))%Z in
	      let v' := rat_to_stream (y * c)%Z d in
		series_body _ arctan_1_n_aux (v+v')%S (2*y, (i+1), n, a+1)%Z.

Definition arctan_1_n (n:Z): stream idigit :=
  arctan_1_n_aux (rat_to_stream  (3*n^2-1) (3*n^3)) (1, 1, n, 0).

Definition pi_div_4 :=
  arctan_1_n 2 + arctan_1_n 3.

(*
(* legere optimisation encore que...*)
/////


Definition terme_arctan_1_n (a n_4ip1 n_2: Z) := 
  (n_2*(a+2)-a, n_2*n_4ip1*a*(a+2))%Z.

CoFixpoint arctan_1_n_aux (v:stream idigit)(s :Z*Z*Z*Z*Z) : stream idigit :=
  match s with (y, a, n_2, n_4, n_4ip1) =>
    let (c, d) :=  terme_arctan_1_n a n_4ip1 n_2 in
      let v' := rat_to_stream (y * c)%Z d in
	series_body _ arctan_1_n_aux (v+v')%S (2*y, (a+4), n_2, n_4, n_4ip1 * n_4)%Z
end.

Definition arctan_1_n (n:Z): stream idigit :=
  let sqr_n := (n*n)%Z in
    arctan_1_n_aux (rat_to_stream (3*n*n-1) (3*n*n*n))
      (1, 5, sqr_n, (sqr_n*sqr_n)%Z, (sqr_n*sqr_n*n)%Z).


*)
