Require Export Reals.
Require Export Field.
Require Export ZArith.
Require Export Omega.
Require Export ZArithRing.
Require Export Wellfounded.
Require Export Zwf.
Require Export Fourier.
Require Export List.

Open Scope Z_scope.

CoInductive stream (A:Set) : Set :=
  Cons : A -> stream A -> stream A.


Implicit Arguments Cons.

Infix "::" := Cons :stream_scope.

Open Scope stream_scope.

Inductive idigit (b:Z)  : Set := 
  | I : Z -> idigit b.

Definition bad_idigit : idigit 9 := I 9 4.

CoInductive represents (b :Z): stream (idigit b) -> R -> Prop :=
  | rep : forall s r k, represents b s r ->
    (-1 <= r <= 1)%R -> 
    -2*b+1 <= k <= 2*b-1 ->
    represents _ (Cons (I _ k) s) ((IZR k +r)/ IZR (2*b))%R.
      
CoFixpoint one (b : Z) : stream (idigit b) :=  
  (I _ (2*b-1))::(one _).

CoFixpoint zero (b : Z) : stream (idigit b) :=  
  Cons (I _ 0) (zero _).

CoFixpoint sub_one (b : Z) : stream (idigit b) := 
  Cons (I _ (-2*b+1)) (sub_one _).

(* returns the interval represented by the n first digits of the stream,
   the value is given by (a, b, k) so that the interval actually is
   (a/((2*n)**k), b/((2*n)**k)) *)
Fixpoint bounds (b:Z) (n:nat) (s:stream (idigit b)) {struct n} : Z*Z*Z :=
  match n with
    | 0%nat => (-1,1,0)
    | S p =>
      let (d,tl) := s in
	let (aux,ex) := bounds _ p tl in
	  let (a,c) := aux in
	    match d with
	      I k => (a + k* Zpower (2*b) ex,c + k * Zpower (2*b) ex,ex+1)
	    end
  end.

Definition rts_find_k (b a c :Z) : Z :=
  if (Z_le_gt_dec (-b) (a * b / c)) 
    then 
      if (Z_le_gt_dec (a * b / c) (b-1)) 
	then a * b / c
	else (b-1)
    else (-b).

(* constructs the infinite stream representing a rational number.
   This function works even if the rational number is not between
   0 and 1, but the value is meaningless. *)
CoFixpoint rat_to_stream (b a c:Z) : stream (idigit b):= 
  let k := 2 *
    (if (Z_le_gt_dec 0 c) 
      then rts_find_k b a c 
      else rts_find_k b (-a) (-c))+1  in 
	(Cons (I _ k) (rat_to_stream _ (2*a*b-k*c) c)).


CoFixpoint mult2 (b:Z) (s: stream (idigit b)) : stream (idigit b) :=
  match s with 
    | (I k)::s' => 
      if (Z_le_gt_dec k (-b-1))
	then (sub_one _)
	else if (Z_le_gt_dec (b+1) k)
	  then  (one _)
	  else match s' with
		 | Cons (I k') s'' => 	
		   if (Z_le_gt_dec k' 0)
		     then if (Z_le_gt_dec 0 k')
		       (*k'=0*)
		       then if (Z_le_gt_dec k (-b))
			 then (I _ (-2*b+1))::mult2 _ (I _ (-b)::s'')
			 else if (Z_le_gt_dec b k)
			   then (I _ (2*b-1))::mult2 _ (I _ b::s'')
			   else (I _ (2*k))::mult2 _ s'
		       (*k' <= -1*)
		       else if (Z_le_gt_dec k (-b))
			 then (sub_one _)
			 else (I _ (2*k-1))::mult2 _ (I _ (k'+b)::s'')
		     (*k'>=1*)
		     else if (Z_le_gt_dec b k) 
		       then (one _)
		       else (I _ (2*k+1))::mult2 _ ((I _ (k'-b))::s'')
	       end		 
  end.

CoFixpoint half_sum (b :Z) (x y : stream (idigit b)) : stream (idigit b) :=
  match (x, y) with 
    | ((I k1)::x', (I l1)::y') =>
      if Z_le_gt_dec (Zmod (k1+l1) 2) 0
	then (I _ ((k1+l1)/2))::(half_sum _ x' y')
	else match x' with 
	       | (I k2)::x'' => 
		 if (Z_le_gt_dec k2 (-1))
		   then (I _ ((k1+l1-1)/2))::(half_sum _ ((I _ (k2+2*b))::x'') y')
		   else if (Z_le_gt_dec 1 k2)
		     then (I _ ((k1+l1+1)/2))::(half_sum _ ((I _ (k2-2*b))::x'') y')
		     else match y' with 
			    | (I l2)::y'' => 
			      if (Z_le_gt_dec l2 (-1))
				then (I _ ((k1+l1-1)/2))::
				  (half_sum _ x' ((I _ (l2+2*b))::y''))
				else (I _ ((k1+l1+1)/2))::
				  (half_sum _ ((I _ (k2-b))::x'') ((I _ (l2-b))::y''))
			  end
	     end
  end.

Definition add (b:Z) (x y: stream (idigit b)) : stream (idigit b):=
  mult2 _ (half_sum _ x y).

CoFixpoint Sopp (b :Z) (x: stream (idigit b)) : stream (idigit b) :=
  match x with 
    | (I k)::x' => (I _ (-k))::Sopp _ x'
  end.

CoFixpoint half_sub (b :Z) (x y : stream (idigit b)) : stream (idigit b) :=
  match (x, y) with 
    | ((I k1)::x', (I l1)::y') =>
      if Z_le_gt_dec (Zmod (k1-l1) 2) 0
	then (I _ ((k1-l1)/2))::(half_sub _ x' y')
	else match x' with 
	       | (I k2)::x'' => 
		 if (Z_le_gt_dec k2 (-1))
		   then (I _ ((k1-l1-1)/2))::(half_sub _ ((I _ (k2+2*b))::x'') y')
		   else if (Z_le_gt_dec 1 k2)
		     then (I _ ((k1-l1+1)/2))::(half_sub _ ((I _ (k2-2*b))::x'') y')
		     else match y' with 
			    | (I l2)::y'' => 
			      if (Z_le_gt_dec 1 l2)
				then (I _ ((k1-l1-1)/2))::
				  (half_sub _ x' ((I _ (l2-2*b))::y''))
				else (I _ ((k1-l1+1)/2))::
				  (half_sub _ ((I _ (k2-b))::x'') ((I _ (l2+b))::y''))
			  end
	     end
  end.

Definition sub (b:Z) (x y: stream (idigit b)) : stream (idigit b):=
  mult2 _ (half_sub _ x y).

Definition series_body (b: Z) (A:Set) 
  (series : stream (idigit b) -> A -> stream (idigit b)) 
  (x : stream (idigit b)) (a:A) : stream (idigit b):=
  match x with
    | (I k1)::(I k2)::x' =>
      if (Z_le_gt_dec (-2*b+3) k2)
	then if (Z_le_gt_dec k2 (2*b-3))
	  then (I b k1)::series ((I b k2)::x') a
	  else if (Z_le_gt_dec (2*b-1) k1)
	    then (I b k1)::series ((I b k2)::x') a
	    else (I b (k1+1))::series ((I b (k2-2*b))::x') a
	else if (Z_le_gt_dec k1 (-2*b+1))
	  then (I b k1)::series ((I b k2)::x') a
	  else (I b (k1-1))::series ((I b (k2+2*b))::x') a
  end.

Definition series_body' (b: Z) (A:Set) 
  (series : stream (idigit b) -> A -> stream (idigit b)) 
  (x : stream (idigit b)) (a:A) : stream (idigit b):=
  match x with
    | (I k1)::(I k2)::x' =>
      if (Z_le_gt_dec b k2)
	then if (Z_eq_dec k1 (2*b-1))
	  then (I b k1)::series ((I b k2)::x') a
	  else (I b (k1+1))::series ((I b (k2-2*b))::x') a
	else if (Z_le_gt_dec k2 (-b))
	  then if (Z_eq_dec k1 (-2*b+1))
	    then (I b k1)::series ((I b k2)::x') a
	    else (I b (k1-1))::series ((I b (k2+2*b))::x') a
	  else (I b k1)::series ((I b k2)::x') a
  end. 

CoFixpoint multkdiv2b (b k: Z) (x:stream (idigit b)) : stream (idigit b) :=
  match x with 
    (I l)::x' => (I b l)::multkdiv2b b k x'
  end.

(* compute x + uw *)
CoFixpoint mult_a (b:Z) (x:stream (idigit b)) 
  (p:stream (idigit b) *stream (idigit b)) : stream (idigit b) :=
  let (u,v) := p in
    match u with
      (I k)::u' => series_body b _ (mult_a b) x (u',v)
    end.

Definition mult (b: Z) (x y:stream (idigit b)) : stream (idigit b) :=
  mult2 b (mult2 b (mult_a b (zero b) (x, (I b 0)::(I b 0)::y))).

Inductive triple (b:Z) : Set :=
  mk_triple : stream (idigit b) -> Z -> Z -> (triple b).

CoFixpoint e_series (b:Z) (v:stream (idigit b) )(s :Z*Z*Z) :stream (idigit b) :=
  let (aux, fact_nm1) := s in let (y, n) := aux in
  let (v', n', fact_nm1') := 
     if Z_le_gt_dec (4*b*b*y) (fact_nm1*(n-1))%Z then
        mk_triple b v n fact_nm1
     else
        mk_triple b (add b v (rat_to_stream b y (fact_nm1 * n)))
          (n+1) (fact_nm1*n) in
     series_body  b _ (e_series b) v' (2*b*y, n', fact_nm1')%Z.

Fixpoint e_minus2_init (n: nat) {struct n} : Z*Z := 
  match n with 
    | O => (0, 1)
    | S (S (S O)) => (1, 2)
    | S p => 
      let (p0, factpm1) :=  e_minus2_init p in
	(p0*Z_of_nat p+1, factpm1*Z_of_nat p)%Z
  end.

Definition number_e_minus2 (b:Z): stream (idigit b):=
  let (p, fact2bm1) :=  e_minus2_init (Zabs_nat (2*b)) in 
    e_series _ (rat_to_stream b p fact2bm1) (1, 2*b, fact2bm1)%Z.

Fixpoint stream_to_list (b:Z) (n:nat) (s : stream (idigit b)) {struct n} : list Z :=
  match n with 
    | O => nil 
    | S p => match s with 
	       | (I k)::x' => cons k (stream_to_list b p x')
	     end
  end.

Fixpoint stream_eval (b:Z) (n:nat) (s : stream (idigit b)) {struct n} : Z*Z :=
  match n with 
    | O => (0, 1)
    | S p => 
      match s with 
	| (I k)::x' =>
	  let (r,q) := stream_eval b p x' in 
	    (r+q*k, 2*b*q)%Z
      end
  end.

(*
Infix "+" := add : stream_scope.
Infix "-" := add : stream_scope.
Infix "*" := mult : stream_scope.
Delimit Scope stream_scope with S.
*)



(*
Time Eval compute in (stream_eval _ 30 (number_e_minus2 4)).

CoFixpoint arctan_1_n_aux (v:stream idigit)(s :Z*Z*Z*Z) : stream idigit :=
  let (aux1, a) := s in
    let (aux2, n) := aux1 in
      let (y, i) := aux2 in
	if (Z_le_gt_dec a (4*i-3)) then
	  series_body _ arctan_1_n_aux v (2*y, i, n, a+1)%Z
	else
	  let c := (n^2*(4*i+3)-4*i-1)%Z in 
	    let d := (n^(4*i+3)*(4*i+1)*(4*i+3))%Z in
	      let v' := rat_to_stream (y * c)%Z d in
		series_body _ arctan_1_n_aux (v+v')%S (2*y, (i+1), n, a+1)%Z.

Definition arctan_1_n (n:Z): stream idigit :=
  arctan_1_n_aux (rat_to_stream  (3*n^2-1) (3*n^3)) (1, 1, n, 0).

Definition pi_div_4 :=
  arctan_1_n 2 + arctan_1_n 3.

(*
(* legere optimisation encore que...*)
/////


Definition terme_arctan_1_n (a n_4ip1 n_2: Z) := 
  (n_2*(a+2)-a, n_2*n_4ip1*a*(a+2))%Z.

CoFixpoint arctan_1_n_aux (v:stream idigit)(s :Z*Z*Z*Z*Z) : stream idigit :=
  match s with (y, a, n_2, n_4, n_4ip1) =>
    let (c, d) :=  terme_arctan_1_n a n_4ip1 n_2 in
      let v' := rat_to_stream (y * c)%Z d in
	series_body _ arctan_1_n_aux (v+v')%S (2*y, (a+4), n_2, n_4, n_4ip1 * n_4)%Z
end.

Definition arctan_1_n (n:Z): stream idigit :=
  let sqr_n := (n*n)%Z in
    arctan_1_n_aux (rat_to_stream (3*n*n-1) (3*n*n*n))
      (1, 5, sqr_n, (sqr_n*sqr_n)%Z, (sqr_n*sqr_n*n)%Z).


*)
*)
